"""A data and reward system that does nothing, returning zero reward on every step."""

import logging
from typing import TYPE_CHECKING, Any

from bsk_rl.data.base import Data, DataStore, GlobalReward

logger = logging.getLogger(__name__)


class NoData(Data):
    """Holds no data."""

    def __init__(self, *args, **kwargs):
        """Holds no data."""
        return super().__init__(*args, **kwargs)

    def __add__(self, other):
        """Add nothing to nothing."""
        return self.__class__()


class NoDataStore(DataStore):
    """DataStore for no data."""

    data_type = NoData

    def __init__(self, *args, **kwargs):
        """Stores and generates no data."""
        return super().__init__(*args, **kwargs)

    def compare_log_states(self, old_state, new_state):
        """Always returns no data."""
        return self.data_type()


class NoReward(GlobalReward):
    """GlobalReward for no data."""

    data_store_type = NoDataStore

    def __init__(self, *args, **kwargs):
        """Returns zero reward at every step.

        This reward system is useful for debugging environments, but is not useful for
        training, since reward is always zero for every satellite.
        """
        return super().__init__(*args, **kwargs)

    def calculate_reward(self, new_data_dict):
        """Reward nothing."""
        return {sat: 0.0 for sat in new_data_dict.keys()}


__doc_title__ = "No Data"
__all__ = ["NoReward", "NoDataStore", "NoData"]



"""Satellites are the agents in the environment."""

import inspect
import logging
from abc import ABC
from typing import TYPE_CHECKING, Any, Optional, Union
from weakref import proxy

import numpy as np
from Basilisk.utilities import macros
from deprecated import deprecated
from gymnasium import spaces

from bsk_rl.act.actions import select_action_builder
from bsk_rl.obs.observations import ObservationBuilder
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils import vizard
from bsk_rl.utils.functional import (
    AbstractClassProperty,
    Resetable,
    collect_default_args,
    safe_dict_merge,
    valid_func_name,
)
from bsk_rl.utils.orbital import TrajectorySimulator

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.act import Action
    from bsk_rl.data.base import DataStore
    from bsk_rl.obs import Observation
    from bsk_rl.sim import Simulator


SatObs = Any
SatAct = Any


class Satellite(ABC, Resetable):
    """Abstract base class for satellites."""

    dyn_type: type["dyn.DynamicsModel"] = AbstractClassProperty()
    fsw_type: type["fsw.FSWModel"] = AbstractClassProperty()
    observation_spec: list["Observation"] = AbstractClassProperty()
    action_spec: list["Action"] = AbstractClassProperty()

    @classmethod
    def default_sat_args(cls, **kwargs) -> dict[str, Any]:
        """Compile default arguments for :class:`~bsk_rl.sim.dyn.DynamicsModel` and :class:`~bsk_rl.sim.fsw.FSWModel`, replacing those specified.

        Args:
            **kwargs: Arguments to override in the default arguments.

        Returns:
            Dictionary of arguments for simulation models.
        """
        defaults = collect_default_args(cls.dyn_type)
        defaults = safe_dict_merge(defaults, collect_default_args(cls.fsw_type))
        for name in dir(cls.fsw_type):
            if inspect.isclass(getattr(cls.fsw_type, name)) and issubclass(
                getattr(cls.fsw_type, name), fsw.Task
            ):
                defaults = safe_dict_merge(
                    defaults, collect_default_args(getattr(cls.fsw_type, name))
                )

        for k, v in kwargs.items():
            if k not in defaults:
                raise KeyError(f"{k} not a valid key for sat_args")
            defaults[k] = v
        return defaults

    def __init__(
        self,
        name: str,
        sat_args: Optional[dict[str, Any]] = None,
        obs_type=np.ndarray,
        dtype: np.dtype = np.float64,
        variable_interval: bool = True,
    ) -> None:
        """The base satellite agent class.

        Args:
            name: Identifier for satellite; does not need to be unique.
            sat_args: Arguments for :class:`~bsk_rl.sim.dyn.DynamicsModel` and
                :class:`~bsk_rl.sim.fsw.FSWModel` construction. Should be in the form of
                a dictionary with keys corresponding to the arguments of the constructor
                and values that are either the desired value or a function that takes no
                arguments and returns a randomized value.
            obs_type: Observation format for the satellite. The :class:`bsk_rl.obs.observations.ObservationBuilder`
                will convert the observation to this format.
            dtype: Data type for observation np vectors.
            variable_interval: Whether to stop the simulation at terminal events. If
                False, only the ``max_step_duration`` setting in :class:`~bsk_rl.GeneralSatelliteTasking`
                will stop the simulation.
        """
        self.name = name
        self.logger = logging.getLogger(__name__).getChild(self.name)
        if sat_args is None:
            sat_args = self.default_sat_args()
        self.sat_args_generator = self.default_sat_args(**sat_args)
        self.simulator: "Simulator"
        self.fsw: "fsw.FSWModel"
        self.dynamics: "dyn.DynamicsModel"
        self.data_store: "DataStore"
        self.requires_retasking: bool
        self.variable_interval = variable_interval
        self._timed_terminal_event_name = None
        self.observation_builder = ObservationBuilder(
            self, obs_type=obs_type, dtype=dtype
        )
        self.action_builder = select_action_builder(self)

    @property
    @deprecated(reason="Use satellite.name instead")
    def id(self) -> str:
        """Unique human-readable identifier."""
        return self.name

    def generate_sat_args(self, **kwargs) -> None:
        """Instantiate sat_args from any randomizers in provided sat_args.

        Args:
            **kwargs: Arguments to override in the default arguments.
        """
        self.sat_args = {
            k: v if not callable(v) else v() for k, v in self.sat_args_generator.items()
        }
        for k, v in kwargs.items():
            if k not in self.sat_args:
                raise KeyError(f"{k} not a valid key for sat_args")
            if np.any(self.sat_args[k] != v):
                self.logger.debug(
                    f"Overwriting {k}={self.sat_args[k]} in sat_args with {v}"
                )
            self.sat_args[k] = v

        self.logger.debug(f"Satellite initialized with {self.sat_args}")

    def reset_overwrite_previous(self) -> None:
        """Overwrite attributes from previous episode."""
        self.requires_retasking = True
        self._timed_terminal_event_name = None
        self._is_alive = True
        self.time_of_death = None
        self.observation_builder.reset_overwrite_previous()
        self.action_builder.reset_overwrite_previous()

    @vizard.visualize
    def create_vizard_data(self, color, vizSupport=None) -> None:
        """Create a location to store data to be passed to enableUnityVisualization."""
        self.vizard_color = color
        self.vizard_data = dict(
            spriteList=vizSupport.setSprite("SQUARE", color=color),
        )

    def reset_pre_sim_init(self) -> None:
        """Called during environment reset, before Basilisk simulation initialization."""
        self.trajectory = TrajectorySimulator(
            utc_init=self.sat_args["utc_init"],
            rN=self.sat_args["rN"],
            vN=self.sat_args["vN"],
            oe=self.sat_args["oe"],
            mu=self.sat_args["mu"],
        )
        self.observation_builder.reset_pre_sim_init()
        self.action_builder.reset_pre_sim_init()

    def set_simulator(self, simulator: "Simulator"):
        """Set the simulator for models.

        Called during simulator initialization.

        Args:
            simulator: Basilisk simulator

        :meta private:
        """
        self.simulator = proxy(simulator)

    def set_dynamics(self, dyn_rate: float) -> "dyn.DynamicsModel":
        """Create dynamics model; called during simulator initialization.

        Args:
            dyn_rate: rate for dynamics simulation [s]

        Returns:
            Satellite's dynamics model

        :meta private:
        """
        dynamics = self.dyn_type(self, dyn_rate, **self.sat_args)
        self.dynamics = proxy(dynamics)
        return dynamics

    def set_fsw(self, fsw_rate: float) -> "fsw.FSWModel":
        """Create flight software model; called during simulator initialization.

        Args:
            fsw_rate: rate for FSW simulation [s]

        Returns:
            Satellite's FSW model

        :meta private:
        """
        fsw = self.fsw_type(self, fsw_rate, **self.sat_args)
        self.fsw = proxy(fsw)
        return fsw

    def reset_during_sim_init(self) -> None:
        """Called during environment reset, during Basilisk simulation initialization."""
        self.observation_builder.reset_during_sim_init()
        self.action_builder.reset_during_sim_init()
        return super().reset_during_sim_init()

    def reset_post_sim_init(self) -> None:
        """Called during environment reset, after Basilisk simulation initialization."""
        self.observation_builder.reset_post_sim_init()
        self.action_builder.reset_post_sim_init()

    @property
    def observation_space(self) -> spaces.Space:
        """Observation space for single satellite, determined from observation.

        Returns:
            gymanisium observation space
        """
        return self.observation_builder.observation_space

    def get_obs(self) -> SatObs:
        """Construct the satellite's observation.

        Returns:
            satellite observation
        """
        return self.observation_builder.get_obs()

    @property
    def observation_description(self) -> Any:
        """Human-interpretable description of observation space."""
        return self.observation_builder.observation_description

    @property
    def action_space(self) -> spaces.Space:
        """Action space for single satellite.

        Returns:
            gymanisium action space
        """
        return self.action_builder.action_space

    def set_action(self, action: Any) -> None:
        """Enable certain processes in the simulator to command the satellite task.

        Args:
            action: Action to take, according to the :class:`action_spec`
        """
        self.action_builder.set_action(action)

    @property
    def action_description(self) -> Any:
        """Human-interpretable description of action space."""
        return self.action_builder.action_description

    def is_alive(self, log_failure=True) -> bool:
        """Check if the satellite is violating any aliveness requirements.

        Checks aliveness checkers in dynamics and FSW models.

        Returns:
            is_alive
        """
        if not self._is_alive:
            return False

        self._is_alive = self.dynamics.is_alive(
            log_failure=log_failure
        ) and self.fsw.is_alive(log_failure=log_failure)
        if not self._is_alive:
            self.record_death(self.simulator.sim_time)
        return self._is_alive

    def record_death(self, time: float) -> None:
        """Record the time of death of the satellite, if not already recorded.

        Args:
            time: Time of death [s]
        """
        if self.time_of_death is None:
            self.time_of_death = time

    @property
    def _satellite_command(self) -> str:
        """Generate string that refers to self in simBase."""
        return f"self.get_satellite('{self.name}')"

    def _info_command(self, info: str) -> str:
        """Generate command to log to info from an event.

        Args:
            info: information to log; cannot include `'` or `"`

        Returns:
            actionList action for simBase.createNewEvent
        """
        return self._satellite_command + f".logger.info('{info}')"

    @deprecated(reason="Use satellite.logger.info instead")
    def log_info(self, info: Any) -> None:
        """Record information at the current simulation time.

        :meta private:

        Args:
            info: Information to log
        """
        self.logger.info(f"{info}")

    @deprecated(reason="Use satellite.logger.warning instead")
    def log_warning(self, warning: Any) -> None:
        """Record warning at the current simulation time.

        :meta private:

        Args:
            warning: Warning to log
        """
        self.logger.warning(f"{warning}")

    def update_timed_terminal_event(
        self,
        t_close: float,
        info: str = "",
        extra_actions: Optional[Union[list[str], callable]] = None,
    ) -> None:
        """Create a simulator event that stops the simulation a certain time.

        Args:
            t_close: Termination time [s]
            info: Additional identifying info to log at terminal time
            extra_actions: Additional actions to perform at terminal time
        """
        self.disable_timed_terminal_event()
        self.logger.info(f"setting timed terminal event at {t_close:.1f}")

        # Create new timed terminal event
        self._timed_terminal_event_name = valid_func_name(
            f"timed_terminal_{t_close}_{self.name}"
        )

        def side_effect(sim):
            self.logger.info(f"timed termination at {t_close:.1f} " + info)
            self.requires_retasking = True
            if extra_actions is not None:
                if callable(extra_actions):
                    extra_actions(sim)
                elif isinstance(extra_actions, list):
                    for action in extra_actions:
                        exec(action, {"self": sim})

        self.simulator.createNewEvent(
            self._timed_terminal_event_name,
            macros.sec2nano(self.simulator.sim_rate),
            True,
            conditionTime=macros.sec2nano(t_close),
            actionFunction=side_effect,
            terminal=self.variable_interval,
        )
        self.simulator.eventMap[self._timed_terminal_event_name].eventActive = True

    def disable_timed_terminal_event(self) -> None:
        """Turn off simulator termination due to :class:`update_timed_terminal_event`."""
        if (
            self._timed_terminal_event_name is not None
            and self._timed_terminal_event_name in self.simulator.eventMap
        ):
            self.simulator.delete_event(self._timed_terminal_event_name)



"""Target scenarios distribute ground targets with some distribution.

Currently, targets are all known to the satellites a priori and are available based on
the imaging requirements given by the dynamics and flight software models.
"""

import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Iterable, Optional, Union

import numpy as np
import pandas as pd
from Basilisk.utilities import orbitalMotion

from bsk_rl.scene import Scenario
from bsk_rl.utils import vizard
from bsk_rl.utils.orbital import lla2ecef

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.data.base import Data
    from bsk_rl.sats import Satellite

logger = logging.getLogger(__name__)


class Target:
    """Ground target with associated value."""

    def __init__(self, name: str, r_LP_P: Iterable[float], priority: float) -> None:
        """Ground target with associated priority and location.

        Args:
            name: Identifier; does not need to be unique
            r_LP_P: Planet-fixed, planet relative location [m]
            priority: Value metric.
        """
        self.name = name
        self.r_LP_P = np.array(r_LP_P)
        self.priority = priority

    @property
    def id(self) -> str:
        """Get unique, human-readable identifier."""
        try:
            return self._id
        except AttributeError:
            self._id = f"{self.name}_{id(self)}"
            return self._id

    def __repr__(self) -> str:
        """Get string representation of target.

        Use ``target.id`` for a unique string identifier.

        Returns:
            Target string
        """
        return f"Target({self.name})"


class UniformTargets(Scenario):
    """Environment with targets distributed uniformly."""

    def __init__(
        self,
        n_targets: Union[int, tuple[int, int]],
        priority_distribution: Optional[Callable] = None,
        radius: float = orbitalMotion.REQ_EARTH * 1e3,
    ) -> None:
        """An environment with evenly-distributed static targets.

        Can be used with :class:`~bsk_rl.data.UniqueImageReward`.

        Args:
            n_targets: Number of targets to generate. Can also be specified as a range
                ``(low, high)`` where the number of targets generated is uniformly selected
                ``low ≤ n_targets ≤ high``.
            priority_distribution: Function for generating target priority. Defaults
                to ``lambda: uniform(0, 1)`` if not specified.
            radius: [m] Radius to place targets from body center. Defaults to Earth's
                equatorial radius.
        """
        self._n_targets = n_targets
        if priority_distribution is None:
            priority_distribution = lambda: np.random.rand()  # noqa: E731
        self.priority_distribution = priority_distribution
        self.radius = radius

    def reset_overwrite_previous(self) -> None:
        """Overwrite target list from previous episode."""
        self.targets = []

    def reset_pre_sim_init(self) -> None:
        """Regenerate target set for new episode."""
        if isinstance(self._n_targets, int):
            self.n_targets = self._n_targets
        else:
            self.n_targets = np.random.randint(self._n_targets[0], self._n_targets[1])
        logger.info(f"Generating {self.n_targets} targets")
        self.regenerate_targets()
        for satellite in self.satellites:
            if hasattr(satellite, "add_location_for_access_checking"):
                for target in self.targets:
                    satellite.add_location_for_access_checking(
                        object=target,
                        r_LP_P=target.r_LP_P,
                        min_elev=satellite.sat_args_generator[
                            "imageTargetMinimumElevation"
                        ],  # Assume not randomized
                        type="target",
                    )

    def reset_during_sim_init(self) -> None:
        """Visualize targets in Vizard on reset."""
        for target in self.targets:
            self.visualize_target(target)

    @vizard.visualize
    def visualize_target(self, target, vizSupport=None, vizInstance=None):
        """Visualize target in Vizard."""
        vizSupport.addLocation(
            vizInstance,
            stationName=target.name,
            parentBodyName="earth",
            r_GP_P=list(target.r_LP_P),
            fieldOfView=np.arctan(500 / 800),
            color=vizSupport.toRGBA255("white"),
            range=1000.0 * 1000,  # meters
            markerScale=np.sqrt(target.priority),
        )
        if vizInstance.settings.showLocationCones == 0:
            vizInstance.settings.showLocationCones = -1
        if vizInstance.settings.showLocationCommLines == 0:
            vizInstance.settings.showLocationCommLines = -1
        if vizInstance.settings.showLocationLabels == 0:
            vizInstance.settings.showLocationLabels = -1

    def regenerate_targets(self) -> None:
        """Regenerate targets uniformly.

        Override this method (as demonstrated in :class:`CityTargets`) to generate
        other distributions.
        """
        self.targets = []
        for i in range(self.n_targets):
            x = np.random.normal(size=3)
            x *= self.radius / np.linalg.norm(x)
            self.targets.append(
                Target(name=f"tgt-{i}", r_LP_P=x, priority=self.priority_distribution())
            )


class CityTargets(UniformTargets):
    """Environment with targets distributed around population centers."""

    def __init__(
        self,
        n_targets: Union[int, tuple[int, int]],
        n_select_from: Optional[int] = None,
        location_offset: float = 0,
        priority_distribution: Optional[Callable] = None,
        radius: float = orbitalMotion.REQ_EARTH * 1e3,
    ) -> None:
        """Construct environment with static targets around population centers.

        Uses the `simplemaps Word Cities Database <https://simplemaps.com/data/world-cities>`_
        for population center locations.

        Args:
            n_targets: Number of targets to generate, as a fixed number or a range.
            n_select_from: Generate targets from the top `n_select_from` most populous
                cities. Will use all cities in the database if not specified.
            location_offset: [m] Offset targets randomly from the city center by up to
                this amount.
            priority_distribution: Function for generating target priority.
            radius: Radius to place targets from body center.
        """
        super().__init__(n_targets, priority_distribution, radius)
        if n_select_from == "all" or n_select_from is None:
            n_select_from = sys.maxsize
        self.n_select_from = n_select_from
        self.location_offset = location_offset

    def regenerate_targets(self) -> None:
        """Regenerate targets based on cities.

        :meta private:
        """
        self.targets = []
        cities = pd.read_csv(
            Path(os.path.realpath(__file__)).parent.parent
            / "_dat"
            / "simplemaps_worldcities"
            / "worldcities.csv",
        )

        if self.n_select_from > len(cities):
            self.n_select_from = len(cities)

        for i in np.random.choice(self.n_select_from, self.n_targets, replace=False):
            city = cities.iloc[i]
            location = lla2ecef(city["lat"], city["lng"], self.radius)
            offset = np.random.normal(size=3)
            offset /= np.linalg.norm(offset)
            offset *= self.location_offset
            location += offset
            location /= np.linalg.norm(location)
            location *= self.radius
            self.targets.append(
                Target(
                    name=f"{city['city']}, {city['iso2']}".replace("'", ""),
                    r_LP_P=location,
                    priority=self.priority_distribution(),
                )
            )


__doc_title__ = "Target Scenarios"
__all__ = ["Target", "UniformTargets", "CityTargets"]



"""Discrete actions are indexable by integer."""

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Union

import numpy as np
from gymnasium import spaces

from bsk_rl.act.actions import Action, ActionBuilder

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sats import Satellite
    from bsk_rl.scene.targets import Target

logger = logging.getLogger(__name__)


class DiscreteActionBuilder(ActionBuilder):
    def __init__(self, satellite: "Satellite") -> None:
        """Processes actions for a discrete action space.

        Args:
            satellite: Satellite to create actions for.
        """
        super().__init__(satellite)
        self.prev_action_key = None

    def reset_post_sim_init(self) -> None:
        """Log previous action key."""
        super().reset_post_sim_init()
        self.prev_action_key = None

    @property
    def action_space(self) -> spaces.Discrete:
        """Discrete action space."""
        return spaces.Discrete(sum([act.n_actions for act in self.action_spec]))

    @property
    def action_description(self) -> list[str]:
        """Return a list of strings corresponding to action names."""
        actions = []
        for act in self.action_spec:
            if act.n_actions == 1:
                actions.append(act.name)
            else:
                actions.extend([f"{act.name}_{i}" for i in range(act.n_actions)])
        return actions

    def set_action(self, action: int) -> None:
        """Sets the action based on the integer index.

        If the action is not an integer, the satellite will attempt to call ``set_action_override``
        for each action, in order, until one works.
        """
        self.satellite.disable_timed_terminal_event()
        if not np.issubdtype(type(action), np.integer):
            logger.warning(
                f"Action '{action}' is not an integer. Will attempt to use compatible set_action_override method."
            )
            for act in self.action_spec:
                try:
                    self.prev_action_key = act.set_action_override(
                        action, prev_action_key=self.prev_action_key
                    )
                    return
                except AttributeError:
                    pass
                except TypeError:
                    pass
            else:
                raise ValueError(
                    f"Action '{action}' is not an integer and no compatible set_action_override method found."
                )
        index = 0
        for act in self.action_spec:
            if index + act.n_actions > action:
                self.prev_action_key = act.set_action(
                    action - index, prev_action_key=self.prev_action_key
                )
                return
            index += act.n_actions
        else:
            raise ValueError(f"Action index {action} out of range.")


class DiscreteAction(Action):
    builder_type = DiscreteActionBuilder

    def __init__(self, name: str = "discrete_act", n_actions: int = 1):
        """Base class for discrete, integer-indexable actions.

        A discrete action may represent multiple indexed actions of the same type.

        Optionally, discrete actions may have a ``set_action_override`` function defined.
        If the action passed to the satellite is not an integer, the satellite will iterate
        over the ``action_spec`` and attempt to call ``set_action_override`` on each action
        until one is successful.

        Args:
            name: Name of the action.
            n_actions: Number of actions available.
        """
        super().__init__(name=name)
        self.n_actions = n_actions

    @abstractmethod
    def set_action(self, action: int, prev_action_key=None) -> str:
        """Activate an action by local index."""
        pass


class DiscreteFSWAction(DiscreteAction):
    def __init__(
        self,
        fsw_action,
        name=None,
        duration: Optional[float] = None,
        reset_task: bool = False,
    ):
        """Discrete action to task a flight software action function.

        This action executes a function of a :class:`~bsk_rl.sim.fsw.FSWModel`
        instance that takes no arguments, typically decorated with ``@action``.

        Args:
            fsw_action: Name of the flight software function to task.
            name: Name of the action. If not specified, defaults to the ``fsw_action`` name.
            duration: Duration of the action in seconds. Defaults to a large value so that
                the :class:`~bsk_rl.GeneralSatelliteTasking` ``max_step_duration``
                controls step length.
            reset_task: If true, reset the action if the previous action was the same.
                Generally, this parameter should be false to ensure realistic, continuous
                operation of satellite modes; however, some Basilisk modules may require
                frequent resetting for normal operation.
        """
        if name is None:
            name = fsw_action
        super().__init__(name=name, n_actions=1)
        self.fsw_action = fsw_action
        self.reset_task = reset_task
        if duration is None:
            duration = 1e9
        self.duration = duration

    def set_action(self, action: int, prev_action_key=None) -> str:
        """Activate the ``fsw_action`` function.

        Args:
            action: Should always be ``1``.
            prev_action_key: Previous action key.

        Returns:
            The name of the activated action.
        """
        assert action == 0
        self.satellite.logger.info(f"{self.name} tasked for {self.duration} seconds")
        self.satellite.update_timed_terminal_event(
            self.simulator.sim_time + self.duration, info=f"for {self.fsw_action}"
        )

        if self.reset_task or prev_action_key != self.fsw_action:
            getattr(self.satellite.fsw, self.fsw_action)()

        return self.fsw_action


class Charge(DiscreteFSWAction):
    def __init__(self, name: Optional[str] = None, duration: Optional[float] = None):
        """Action to enter a sun-pointing charging mode (:class:`~bsk_rl.sim.fsw.BasicFSWModel.action_charge`).

        Charging will only occur if the satellite is in sunlight.

        Args:
            name: Action name.
            duration: Time to task action, in seconds.
        """
        super().__init__(fsw_action="action_charge", name=name, duration=duration)


class Drift(DiscreteFSWAction):
    def __init__(self, name: Optional[str] = None, duration: Optional[float] = None):
        """Action to disable all FSW tasks (:class:`~bsk_rl.sim.fsw.BasicFSWModel.action_drift`).

        Args:
            name: Action name.
            duration: Time to task action, in seconds.
        """
        super().__init__(fsw_action="action_drift", name=name, duration=duration)


class Desat(DiscreteFSWAction):
    def __init__(self, name: Optional[str] = None, duration: Optional[float] = None):
        """Action to desaturate reaction wheels (:class:`~bsk_rl.sim.fsw.BasicFSWModel.action_desat`).

        This action must be called repeatedly to fully desaturate the reaction wheels.

        Args:
            name: Action name.
            duration: Time to task action, in seconds.
        """
        super().__init__(
            fsw_action="action_desat", name=name, duration=duration, reset_task=True
        )


class Downlink(DiscreteFSWAction):
    def __init__(self, name: Optional[str] = None, duration: Optional[float] = None):
        """Action to transmit data from the data buffer (:class:`~bsk_rl.sim.fsw.ImagingFSWModel.action_downlink`).

        If not in range of a ground station (defined in
        :class:`~bsk_rl.sim.world.GroundStationWorldModel`), no data will
        be downlinked.

        Args:
            name: Action name.
            duration: Time to task action, in seconds.
        """
        super().__init__(fsw_action="action_downlink", name=name, duration=duration)


class Scan(DiscreteFSWAction):
    def __init__(self, name: Optional[str] = None, duration: Optional[float] = None):
        """Action to collect data from a :class:`~bsk_rl.scene.UniformNadirScanning` (:class:`~bsk_rl.sim.fsw.ContinuousImagingFSWModel.action_nadir_scan`).

        Args:
            name: Action name.
            duration: Time to task action, in seconds.
        """
        super().__init__(fsw_action="action_nadir_scan", name=name, duration=duration)


class Image(DiscreteAction):
    def __init__(
        self,
        n_ahead_image: int,
        max_duration: Optional[float] = None,
        name: str = "action_image",
    ):
        """Actions to image upcoming target (:class:`~bsk_rl.sim.fsw.ImagingFSWModel.action_image`).

        Adds ``n_ahead_image`` actions to the action space, corresponding to the next
        ``n_ahead_image`` unimaged targets. The action may be unsuccessful if the target
        exits the satellite's field of regard before the satellite settles on the target
        and takes an image. The action with stop as soon as the image is successfully
        taken, or when the the target exits the field of regard.

        This action implements a ``set_action_override`` that allows a target to be tasked
        based on the target's ID string or the Target object.

        Args:
            n_ahead_image: Number of unimaged, along-track targets to consider.
            max_duration: Maximum time to task the action, in seconds. If ``None``, tasks until
                the end of the next opportunity.
            name: Action name.
        """
        from bsk_rl.sats import ImagingSatellite

        self.satellite: "ImagingSatellite"
        self.max_duration = max_duration
        super().__init__(name=name, n_actions=n_ahead_image)

    def image(
        self, target: Union[int, "Target", str], prev_action_key: Optional[str] = None
    ) -> str:
        """Task or retask a satellite for imaging a target.

        Args:
            target: Target to image.
            prev_action_key: Previous action key.

        :meta private:
        """
        target = self.satellite.parse_target_selection(target)
        if target.id != prev_action_key:
            self.satellite.task_target_for_imaging(
                target, max_duration=self.max_duration
            )
        else:
            self.satellite.enable_target_window(target, max_duration=self.max_duration)

        return target.id

    def set_action(self, action: int, prev_action_key: Optional[str] = None) -> str:
        """Image a target by local index.

        Args:
            action: Index of the target to image.
            prev_action_key: Previous action key.

        :meta_private:
        """
        self.satellite.logger.info(f"target index {action} tasked")
        return self.image(action, prev_action_key)

    def set_action_override(
        self, action: Union["Target", str], prev_action_key: Optional[str] = None
    ) -> str:
        """Image a target by target index, Target, or ID.

        Args:
            action: Target to image in the form of a Target object, target ID, or target index.
            prev_action_key: Previous action key.

        :meta_private:
        """
        return self.image(action, prev_action_key)


class Broadcast(DiscreteAction):
    def __init__(
        self,
        name: str = "action_broadcast",
        duration: Optional[float] = None,
    ):
        """Action to broadcast data to all satellites in the simulation.

        Should be used with a :class:`~bsk_rl.comm.BroadcastCommunication`-derived
        communication method to limit communication to broadcasting satellites.

        Args:
            name: Action name.
            duration: [s] Time to idle before communicating.
        """
        super().__init__(name=name, n_actions=1)
        if duration is None:
            duration = 1e9
        self.duration = duration

    def reset_post_sim_init(self) -> None:
        """Log previous action key."""
        super().reset_post_sim_init()
        self.broadcast_pending = False

    def set_action(self, action: int, prev_action_key=None) -> str:
        assert action == 0
        self.broadcast_pending = True
        self.satellite.update_timed_terminal_event(
            self.simulator.sim_time + self.duration, info=f"for broadcast"
        )
        return self.name


__doc_title__ = "Discrete Backend"
__all__ = ["DiscreteActionBuilder"]



"""``bsk_rl.utils.rllib.discounting`` is a collection of utilities for semi-MDP style discounting.

See the following examples for how to use these utilities:

* :doc:`/examples/time_discounted_gae` - An example of :class:`TimeDiscountedGAEPPOTorchLearner`
  in a single-agent case.
* :doc:`/examples/async_multiagent_training` - An example of the time discounted learner
  and connectors for asynchronous multi-agent training (:class:`ContinuePreviousAction`,
  :class:`MakeAddedStepActionValid`, and :class:`CondenseMultiStepActions`).
"""

from logging import getLogger
from typing import Any, List, Literal, Optional

import numpy as np
from ray.rllib.algorithms.ppo.ppo_learner import PPOLearner
from ray.rllib.algorithms.ppo.tf.ppo_tf_learner import PPOTfLearner
from ray.rllib.algorithms.ppo.torch.ppo_torch_learner import PPOTorchLearner
from ray.rllib.connectors.connector_v2 import ConnectorV2
from ray.rllib.core.columns import Columns
from ray.rllib.evaluation.postprocessing import Postprocessing
from ray.rllib.utils.numpy import convert_to_numpy
from ray.rllib.utils.postprocessing.episodes import (
    add_one_ts_to_episodes_and_truncate,
    remove_last_ts_from_data,
    remove_last_ts_from_episodes_and_restore_truncateds,
)
from ray.rllib.utils.postprocessing.zero_padding import unpad_data_if_necessary
from ray.rllib.utils.typing import EpisodeType

from bsk_rl import NO_ACTION
from bsk_rl.gym import is_no_action, no_action_like

logger = getLogger(__name__)


class ContinuePreviousAction(ConnectorV2):
    def __init__(self, *args, **kwargs):
        """Override actions with ``NO_ACTION`` on connector pass if the agent does not require retasking."""
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        *,
        data: Optional[Any],
        episodes: List[EpisodeType],
        **_,
    ) -> Any:
        """Override actions with ``NO_ACTION`` on connector pass.

        :meta private:
        """
        for sa_episode in self.single_agent_episode_iterator(
            episodes,
            agents_that_stepped_only=True,
        ):
            if not sa_episode.get_infos(-1)["requires_retasking"]:
                if sa_episode.agent_id is None:
                    assert len(data[Columns.ACTIONS]) == 1
                    id_tuple = list(data[Columns.ACTIONS].keys())[0]
                else:
                    id_tuples = [
                        id_tuple
                        for id_tuple in data[Columns.ACTIONS].keys()
                        if id_tuple[1] == sa_episode.agent_id
                    ]
                    if len(id_tuples) == 0:
                        continue
                    else:
                        id_tuple = id_tuples[0]

                data[Columns.ACTIONS][id_tuple][0] = (
                    data[Columns.ACTIONS][id_tuple][0] * 0 + NO_ACTION
                )

        return data


class ContinuePreviousActionAppended(ConnectorV2):
    def __init__(self, *args, **kwargs):
        """Override actions with ``NO_ACTION`` on connector pass if the agent does not require retasking.

        This additional connector must be appended after NormalizeAndClipActions in continuous
        action spaces. For example:

        .. code-block:: python

            # Append extra connector to the pipeline
            old_connector_builder = config.build_module_to_env_connector

            def new_connector_builder(env):
                pipeline = old_connector_builder(env)
                after = pipeline.insert_after(
                    "NormalizeAndClipActions", ContinuePreviousActionAppended()
                )
                return pipeline

            config.build_module_to_env_connector = new_connector_builder

        """
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        *,
        data: Optional[Any],
        episodes: List[EpisodeType],
        **_,
    ) -> Any:
        """Override actions with ``NO_ACTION`` on connector pass.

        :meta private:
        """
        for id_tuple in data[Columns.ACTIONS_FOR_ENV].keys():
            if is_no_action(data[Columns.ACTIONS][id_tuple][0]):
                data[Columns.ACTIONS_FOR_ENV][id_tuple][0] = no_action_like(
                    data[Columns.ACTIONS_FOR_ENV][id_tuple][0]
                )

        return data


class MakeAddedStepActionValid(ConnectorV2):
    def __init__(self, *args, expected_train_batch_size, **kwargs):
        """Ensure that padded steps are not duplicates of ``NO_ACTION`` steps."""
        super().__init__(*args, **kwargs)
        self.expected_train_batch_size = expected_train_batch_size

    def __call__(
        self,
        *,
        data: Optional[Any],
        episodes: List[EpisodeType],
        **_,
    ) -> Any:
        """Ensure that padded steps are not duplicates of ``NO_ACTION`` steps.

        :meta private:
        """
        total_episodes = 0
        total_steps = 0
        episode_lens = []
        episode_multi_agent_ids = []
        for episode in self.single_agent_episode_iterator(
            episodes, agents_that_stepped_only=False
        ):
            episode_lens.append(len(episode))
            episode_multi_agent_ids.append(episode.multi_agent_episode_id)

        if episode_multi_agent_ids[0] is None:
            total_episodes = len(episode_lens)
            total_steps = sum(episode_lens)
        else:
            total_episodes = len(set(episode_multi_agent_ids))
            max_lens = {}
            for episode_id, length in zip(episode_multi_agent_ids, episode_lens):
                if episode_id not in max_lens or length > max_lens[episode_id]:
                    max_lens[episode_id] = length
            total_steps = sum(max_lens.values())

        one_ts_added = False
        if total_steps == self.expected_train_batch_size + total_episodes:
            one_ts_added = True

        if one_ts_added:
            for episode in self.single_agent_episode_iterator(
                episodes, agents_that_stepped_only=False
            ):
                # Arbitrary filler, will be deleted by remove_last_ts_from_episodes_and_restore_truncateds
                if is_no_action(episode.actions[-1]):
                    episode.actions[-1] = episode._action_space.sample()
                episode.validate()

        return data


class CondenseMultiStepActions(ConnectorV2):
    def __init__(self, *args, **kwargs):
        """Combine steps that used ``NO_ACTION`` on connector pass."""
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        *,
        data: Optional[Any],
        episodes: List[EpisodeType],
        **_,
    ) -> Any:
        """Combine steps that used ``NO_ACTION`` on connector pass.

        :meta private:
        """
        for episode in self.single_agent_episode_iterator(
            episodes, agents_that_stepped_only=False
        ):
            action_idx = list(
                np.argwhere(
                    [not is_no_action(action) for action in episode.actions.data]
                ).flatten()
            )
            obs_idx = action_idx.copy()
            obs_idx.append(len(episode) - 1)

            lookback = episode.actions.data[: episode.actions.lookback]
            new_lookback = episode.actions.lookback
            for action in lookback:
                if is_no_action(action):
                    new_lookback -= 1
                else:
                    break

            # Only keep non-None actions
            episode.actions.data = np.array(episode.actions.data)[action_idx]
            episode.actions.lookback = new_lookback
            for column in episode.extra_model_outputs:
                episode.extra_model_outputs[column].data = episode.extra_model_outputs[
                    column
                ].data[action_idx]
                episode.extra_model_outputs[column].lookback = new_lookback

            # Update episode length
            episode.t = episode.t_started + len(episode.actions)

            # Only keep obs that resulted in those actions
            episode.observations.data = np.array(episode.observations.data)[obs_idx]
            episode.observations.lookback = new_lookback

            # Associate subsequent rewards with prior action
            rewards = []
            requires_retasking = []
            for i, idx_start in enumerate(action_idx):
                if i == len(action_idx) - 1:
                    idx_end = len(episode) - 1
                else:
                    idx_end = action_idx[i + 1] - 1
                rewards.append(
                    sum(episode.rewards[idx_start : idx_end + 1])
                )  # Doesn't discount over course of multistep
                requires_retasking.append(
                    episode.infos.data[idx_start]["requires_retasking"]
                )
            requires_retasking.append(True)
            episode.rewards.data = np.array(rewards)
            episode.rewards.lookback = new_lookback

            # Accumulate d_ts from prior action
            d_ts = []
            for i, idx_end in enumerate(obs_idx):
                if i == 0:
                    idx_start = 0
                else:
                    idx_start = obs_idx[i - 1] + 1
                d_ts.append(
                    sum(
                        info["d_ts"]
                        for info in episode.infos.data[idx_start : idx_end + 1]
                    )
                )
            episode.infos.data = [
                dict(d_ts=d_ts, requires_retasking=requires_retasking)
                for d_ts, requires_retasking in zip(d_ts, requires_retasking)
            ]
            episode.infos.lookback = new_lookback

            episode.validate()

        return data


def compute_value_targets_time_discounted(
    values,
    rewards,
    terminateds,
    truncateds,
    step_durations,
    gamma: float,
    lambda_: float,
    reward_time: Literal["step_start", "step_end"] = "step_end",
):
    """Computes value function (vf) targets given vf predictions and rewards.

    Note that advantages can then easily be computed via the formula:
    advantages = targets - vf_predictions
    """
    assert reward_time in ["step_start", "step_end"]

    # Shift step durations to associate with previous timestep
    # delta_t->t+1 comes with t+1's info, but should be used with t
    step_durations = np.concatenate((step_durations[1:], [step_durations[-1]]))

    # Force-set all values at terminals (not at truncations!) to 0.0.
    orig_values = flat_values = values * (1.0 - terminateds)

    flat_values = np.append(flat_values, 0.0)
    if reward_time == "step_start":
        intermediates = (
            rewards + gamma**step_durations * (1 - lambda_) * flat_values[1:]
        )
    elif reward_time == "step_end":
        intermediates = gamma**step_durations * (
            rewards + (1 - lambda_) * flat_values[1:]
        )
    continues = 1.0 - terminateds

    Rs = []
    last = flat_values[-1]
    for t in reversed(range(intermediates.shape[0])):
        last = (
            intermediates[t]
            + continues[t] * gamma ** step_durations[t] * lambda_ * last
        )
        # last = (
        #     intermediates[t]
        #     + continues[t] * gamma * lambda_ * last
        # )

        Rs.append(last)
        if truncateds[t]:
            last = orig_values[t]

    # Reverse back to correct (time) direction.
    value_targets = np.stack(list(reversed(Rs)), axis=0)

    return value_targets.astype(np.float32)


class TimeDiscountedGAEPPOLearner(PPOLearner):
    def __init__(self, *args, **kwargs):
        """Discount episodes according to the ``d_ts`` value in the info dictionary."""
        super().__init__(*args, **kwargs)

    def _compute_gae_from_episodes(
        self,
        *,
        episodes: Optional[List[EpisodeType]] = None,
    ) -> tuple[Optional[dict[str, Any]], Optional[List[EpisodeType]]]:
        if not episodes:
            raise ValueError(
                "`PPOLearner._compute_gae_from_episodes()` must have the `episodes` "
                "arg provided! Otherwise, GAE/advantage computation can't be performed."
            )

        batch = {}

        sa_episodes_list = list(
            self._learner_connector.single_agent_episode_iterator(
                episodes, agents_that_stepped_only=False
            )
        )
        # Make all episodes one ts longer in order to just have a single batch
        # (and distributed forward pass) for both vf predictions AND the bootstrap
        # vf computations.
        orig_truncateds_of_sa_episodes = add_one_ts_to_episodes_and_truncate(
            sa_episodes_list
        )

        # Call the learner connector (on the artificially elongated episodes)
        # in order to get the batch to pass through the module for vf (and
        # bootstrapped vf) computations.
        batch_for_vf = self._learner_connector(
            rl_module=self.module,
            data={},
            episodes=episodes,
            shared_data={},
        )

        # print(batch_for_vf)
        # Perform the value model's forward pass.
        vf_preds = convert_to_numpy(self._compute_values(batch_for_vf))

        for module_id, module_vf_preds in vf_preds.items():
            # Collect new (single-agent) episode lengths.
            episode_lens_plus_1 = [
                len(e)
                for e in sa_episodes_list
                if e.module_id is None or e.module_id == module_id
            ]

            # Remove all zero-padding again, if applicable, for the upcoming
            # GAE computations.
            module_vf_preds = unpad_data_if_necessary(
                episode_lens_plus_1, module_vf_preds
            )

            # Compute value targets.
            reward_time = self.config.learner_config_dict.get("reward_time", "step_end")
            module_value_targets = compute_value_targets_time_discounted(
                values=module_vf_preds,
                rewards=unpad_data_if_necessary(
                    episode_lens_plus_1,
                    convert_to_numpy(batch_for_vf[module_id][Columns.REWARDS]),
                ),
                terminateds=unpad_data_if_necessary(
                    episode_lens_plus_1,
                    convert_to_numpy(batch_for_vf[module_id][Columns.TERMINATEDS]),
                ),
                truncateds=unpad_data_if_necessary(
                    episode_lens_plus_1,
                    convert_to_numpy(batch_for_vf[module_id][Columns.TRUNCATEDS]),
                ),
                step_durations=unpad_data_if_necessary(
                    episode_lens_plus_1,
                    np.array(
                        [
                            info["d_ts"]
                            for info in batch_for_vf[module_id][Columns.INFOS]
                        ]
                    ),
                ),
                gamma=self.config.gamma,
                lambda_=self.config.lambda_,
                reward_time=reward_time,
            )

            # Remove the extra timesteps again from vf_preds and value targets. Now that
            # the GAE computation is done, we don't need this last timestep anymore in
            # any of our data.
            module_vf_preds, module_value_targets = remove_last_ts_from_data(
                episode_lens_plus_1,
                module_vf_preds,
                module_value_targets,
            )
            module_advantages = module_value_targets - module_vf_preds
            # Drop vf-preds, not needed in loss. Note that in the PPORLModule, vf-preds
            # are recomputed with each `forward_train` call anyway.
            # Standardize advantages (used for more stable and better weighted
            # policy gradient computations).
            module_advantages = (module_advantages - module_advantages.mean()) / max(
                1e-4, module_advantages.std()
            )

            # Restructure ADVANTAGES and VALUE_TARGETS in a way that the Learner
            # connector can properly re-batch these new fields.
            batch_pos = 0
            for eps in sa_episodes_list:
                if eps.module_id is not None and eps.module_id != module_id:
                    continue
                len_ = len(eps) - 1
                self._learner_connector.add_n_batch_items(
                    batch=batch,
                    column=Postprocessing.ADVANTAGES,
                    items_to_add=module_advantages[batch_pos : batch_pos + len_],
                    num_items=len_,
                    single_agent_episode=eps,
                )
                self._learner_connector.add_n_batch_items(
                    batch=batch,
                    column=Postprocessing.VALUE_TARGETS,
                    items_to_add=module_value_targets[batch_pos : batch_pos + len_],
                    num_items=len_,
                    single_agent_episode=eps,
                )
                batch_pos += len_

        # Remove the extra (artificial) timesteps again at the end of all episodes.
        remove_last_ts_from_episodes_and_restore_truncateds(
            sa_episodes_list,
            orig_truncateds_of_sa_episodes,
        )

        return batch, episodes


class TimeDiscountedGAEPPOTorchLearner(PPOTorchLearner, TimeDiscountedGAEPPOLearner):
    pass


class TimeDiscountedGAEPPOTfLearner(PPOTfLearner, TimeDiscountedGAEPPOLearner):
    pass


__doc_title__ = "Semi-MDP Discounting in RLlib"
__all__ = [
    "ContinuePreviousAction",
    "MakeAddedStepActionValid",
    "CondenseMultiStepActions",
    "TimeDiscountedGAEPPOTorchLearner",
    "TimeDiscountedGAEPPOTfLearner",
    "TimeDiscountedGAEPPOLearner",
]



from unittest.mock import MagicMock, patch

import numpy as np

from bsk_rl.sim.fsw import BasicFSWModel, FSWModel, ImagingFSWModel, Task, action

module = "bsk_rl.sim.fsw."


def test_action_decorator():
    mock_actions = MagicMock()

    class Object(MagicMock):
        @action
        def action_function(self, foo, bar=2):
            mock_actions(foo, bar=bar)

    o = Object()
    o.tasks = [MagicMock()]
    o.action_function(3, bar=4)
    o.fsw_proc.disableAllTasks.assert_called_once()
    o._zero_gateway_msgs.assert_called_once()
    o.dynamics.reset_for_action.assert_called_once()
    o.tasks[0].reset_for_action.assert_called_once()
    mock_actions.assert_called_once_with(3, bar=4)


@patch.multiple(FSWModel, __abstractmethods__=set())
class TestFSWModel:
    def test_base_class(self):
        sat = MagicMock()
        fsw = FSWModel(sat, 1.0)
        # fsw.simulator.CreateNewProcess.assert_called_once()
        assert sat.simulator.world == fsw.world
        assert sat.dynamics == fsw.dynamics

    @patch(module + "base.check_aliveness_checkers", MagicMock(return_value=True))
    def test_is_alive(self):
        fsw = FSWModel(MagicMock(), 1.0)
        assert fsw.is_alive()


@patch.multiple(Task, __abstractmethods__=set(), name="task")
class TestTask:
    def test_base_class(self):
        fsw = MagicMock()
        task = Task(fsw, 1)
        task.create_task()
        task.fsw.simulator.CreateNewTask.assert_called_once()
        task._setup_fsw_objects()
        task.reset_for_action()
        task.fsw.simulator.disableTask.assert_called_once()


basicfsw = module + "BasicFSWModel."


@patch(basicfsw + "_requires_dyn", MagicMock(return_value=[]))
class TestBasicFSWModel:
    @patch(basicfsw + "_set_messages", MagicMock())
    @patch(basicfsw + "SunPointTask")
    @patch(basicfsw + "NadirPointTask")
    @patch(basicfsw + "RWDesatTask")
    @patch(basicfsw + "TrackingErrorTask")
    @patch(basicfsw + "MRPControlTask")
    def test_make_tasks(self, *args):
        fsw = BasicFSWModel(MagicMock(), 1)
        for task in fsw.tasks:
            task.create_task.assert_called_once()
            task._create_module_data.assert_called_once()
            task._setup_fsw_objects.assert_called_once()


imagingfsw = module + "ImagingFSWModel."


@patch(imagingfsw + "_requires_dyn", MagicMock(return_value=[]))
@patch(imagingfsw + "_make_task_list", MagicMock())
@patch(imagingfsw + "_set_messages", MagicMock())
class TestImagingFSWModel:
    def test_fsw_properties(self):
        fsw = ImagingFSWModel(MagicMock(), 1.0)
        fsw.locPoint = MagicMock(pHat_B=np.array([1.0, 0.0, 0.0]))
        fsw.satellite = MagicMock(dynamics=MagicMock(BP=np.identity(3)))
        assert (fsw.c_hat_P == np.array([1.0, 0.0, 0.0])).all()



"""RSO inspection dynamics models."""

import numpy as np
from Basilisk.simulation import spacecraftLocation
from Basilisk.utilities import macros

from bsk_rl.sim.dyn import BasicDynamicsModel, ContinuousImagingDynModel


class RSODynModel(BasicDynamicsModel):
    """For an RSO with points targets for observation."""

    def __init__(self, *args, **kwargs) -> None:
        """Allow for body fixed inspection points to be added to a spacecraft."""
        super().__init__(*args, **kwargs)

    def _setup_dynamics_objects(self, **kwargs) -> None:
        super()._setup_dynamics_objects(**kwargs)

        rso_dyn_proc_name = "RSODynProcess"
        self.rso_dyn_proc = self.simulator.CreateNewProcess(rso_dyn_proc_name, 1)
        self.rso_task_name = "RSODynTask"
        self.rso_dyn_proc.addTask(
            self.simulator.CreateNewTask(
                self.rso_task_name, macros.sec2nano(self.dyn_rate)
            )
        )

        self.rso_points = []

    def add_rso_point(
        self,
        r_LB_B: np.ndarray,
        aHat_B: np.ndarray,
        theta: float,
        range: float,
        theta_solar: float,
        min_shadow_factor: float,
    ):
        """Add a point on the RSO for observation.

        Args:
            r_LB_B: [m] Position of point in RSO body frame.
            aHat_B: [-] Point normal in RSO body frame.
            theta: [rad] Max angle from instrument to normal.
            range: [m] Max range to point for imaging.
            theta_solar: [rad] Maximum solar incidence angle for illumination.
            min_shadow_factor: [-] Minimum shadow factor for imaging.
        """
        rso_point_model = spacecraftLocation.SpacecraftLocation()
        rso_point_model.primaryScStateInMsg.subscribeTo(self.scObject.scStateOutMsg)

        rso_point_model.planetInMsg.subscribeTo(
            self.world.gravFactory.spiceObject.planetStateOutMsgs[self.world.body_index]
        )
        rso_point_model.rEquator = self.simulator.world.planet.radEquator
        rso_point_model.rPolar = self.simulator.world.planet.radEquator * 0.98

        rso_point_model.sunInMsg.subscribeTo(
            self.world.gravFactory.spiceObject.planetStateOutMsgs[self.world.sun_index]
        )
        rso_point_model.eclipseInMsg.subscribeTo(
            self.world.eclipseObject.eclipseOutMsgs[self.eclipse_index]
        )

        rso_point_model.r_LB_B = r_LB_B
        rso_point_model.aHat_B = aHat_B
        rso_point_model.theta = theta
        rso_point_model.theta_solar = theta_solar
        rso_point_model.min_shadow_factor = min_shadow_factor
        rso_point_model.maximumRange = range
        self.simulator.AddModelToTask(
            self.rso_task_name, rso_point_model, ModelPriority=1
        )

        self.rso_points.append(rso_point_model)
        return rso_point_model


class RSOInspectorDynModel(ContinuousImagingDynModel):
    """For a satellite observing points on an RSO."""

    def __init__(self, *args, **kwargs) -> None:
        """Allow a satellite to observe points in a RSO."""
        super().__init__(*args, **kwargs)

    def add_rso_point(self, rso_point_model):
        """Add a point on the RSO for observation."""
        rso_point_model.addSpacecraftToModel(self.scObject.scStateOutMsg)


__doc_title__ = "RSO Inspection"
__all__ = [
    "RSODynModel",
    "RSOInspectorDynModel",
]



"""``bsk_rl.utils.actuator_primitives``: Preset spacecraft components."""

import numpy as np
from Basilisk.simulation import reactionWheelStateEffector, thrusterDynamicEffector
from Basilisk.utilities import simIncludeRW, simIncludeThruster


def balancedHR16Triad(
    useRandom: bool = False,
    randomBounds: tuple[float, float] = (-400, 400),
    wheelSpeeds: list[float] = [500, 500, 500],
) -> tuple[
    reactionWheelStateEffector.ReactionWheelStateEffector,
    simIncludeRW.rwFactory,
    list[float],
]:
    """Create a set of three HR16 reaction wheels.

    Args:
        useRandom: Use random values for wheel speeds.
        randomBounds: Bounds for random wheel speeds.
        wheelSpeeds: Fixed wheel speeds.

    Returns:
        tuple:
            * **rwStateEffector**: Reaction wheel state effector instance.
            * **rwFactory**: Factory containing defined reaction wheels.
            * **wheelSpeeds**: Wheel speeds.
    """
    rwFactory = simIncludeRW.rwFactory()
    if useRandom:
        wheelSpeeds = np.random.uniform(randomBounds[0], randomBounds[1], 3)

    rwFactory.create(
        "Honeywell_HR16", [1, 0, 0], maxMomentum=50.0, Omega=wheelSpeeds[0]  # RPM
    )
    rwFactory.create(
        "Honeywell_HR16", [0, 1, 0], maxMomentum=50.0, Omega=wheelSpeeds[1]  # RPM
    )
    rwFactory.create(
        "Honeywell_HR16", [0, 0, 1], maxMomentum=50.0, Omega=wheelSpeeds[2]  # RPM
    )

    rwStateEffector = reactionWheelStateEffector.ReactionWheelStateEffector()

    return rwStateEffector, rwFactory, wheelSpeeds


def idealMonarc1Octet() -> tuple:
    """Create a set of eight ADCS thrusters using MOOG Monarc-1 attributes.

    Returns a set of thrusters and thrusterFac instance to add thrusters to a
    spacecraft.

    Returns:
        tuple:
        * **thrusterSet**: Thruster dynamic effector instance.
        * **thrusterFac**: Factory containing defined thrusters.
    """
    location = [
        [3.874945160902288e-2, -1.206182747348013, 0.85245],
        [3.874945160902288e-2, -1.206182747348013, -0.85245],
        [-3.8749451609022656e-2, -1.206182747348013, 0.85245],
        [-3.8749451609022656e-2, -1.206182747348013, -0.85245],
        [-3.874945160902288e-2, 1.206182747348013, 0.85245],
        [-3.874945160902288e-2, 1.206182747348013, -0.85245],
        [3.8749451609022656e-2, 1.206182747348013, 0.85245],
        [3.8749451609022656e-2, 1.206182747348013, -0.85245],
    ]

    direction = [
        [-0.7071067811865476, 0.7071067811865475, 0.0],
        [-0.7071067811865476, 0.7071067811865475, 0.0],
        [0.7071067811865475, 0.7071067811865476, 0.0],
        [0.7071067811865475, 0.7071067811865476, 0.0],
        [0.7071067811865476, -0.7071067811865475, 0.0],
        [0.7071067811865476, -0.7071067811865475, 0.0],
        [-0.7071067811865475, -0.7071067811865476, 0.0],
        [-0.7071067811865475, -0.7071067811865476, 0.0],
    ]
    thrusterSet = thrusterDynamicEffector.ThrusterDynamicEffector()
    thFactory = simIncludeThruster.thrusterFactory()
    for pos_B, dir_B in zip(location, direction):
        thFactory.create("MOOG_Monarc_1", pos_B, dir_B)
    return thrusterSet, thFactory


__doc_title__ = "Actuator Primitives"
__all__ = ["balancedHR16Triad", "idealMonarc1Octet"]



"""Satellites are the agents in the environment."""

import bisect
import logging
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, Union

import numpy as np
from Basilisk.utilities import macros
from scipy.optimize import minimize_scalar, root_scalar

from bsk_rl.sats.satellite import Satellite
from bsk_rl.scene.targets import Target
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils import vizard
from bsk_rl.utils.functional import valid_func_name
from bsk_rl.utils.orbital import elevation

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.data.unique_image_data import UniqueImageStore

logger = logging.getLogger(__name__)

SatObs = Any
SatAct = Any


class AccessSatellite(Satellite):
    """Satellite that detects access opportunities for ground locations."""

    def __init__(
        self,
        *args,
        generation_duration: float = 600.0,
        initial_generation_duration: Optional[float] = None,
        max_generation_duration_beyond_initial: Optional[float] = float("inf"),
        **kwargs,
    ) -> None:
        """Satellite that detects access opportunities for ground locations.

        This satellite can be used to computes access opportunities for ground locations
        such as imaging targets or ground stations. The satellite will calculate upcoming
        opportunities for each location and order the opportunities by close time.
        Opportunities are calculated based on a per-location minimum elevation angle.

        Args:
            args: Passed through to :class:`Satellite` constructor.
            generation_duration: [s] Duration to calculate additional opportunities for
                when the simulation time reaches the current calculation time. If
                `None`, generate opportunities for the simulation `time_limit` unless
                the simulation is infinite.
            initial_generation_duration: [s] Period to calculate opportunities for on
                environment reset.
            max_generation_duration_beyond_initial: [s] Maximum time to calculate opportunities
                beyond the initial generation duration.
            kwargs: Passed through to :class:`Satellite` constructor.
        """
        super().__init__(*args, **kwargs)
        self.generation_duration = generation_duration
        self.initial_generation_duration = initial_generation_duration
        self.max_generation_duration_beyond_initial = (
            max_generation_duration_beyond_initial
        )
        self.access_filter_functions = []
        self.add_access_filter(lambda opportunity: True)

    def reset_overwrite_previous(self) -> None:
        """Overwrite previous opportunities and locations."""
        super().reset_overwrite_previous()
        self.opportunities: list[dict] = []
        self.window_calculation_time = 0
        self.locations_for_access_checking: list[dict[str, Any]] = []

    def add_location_for_access_checking(
        self,
        object: Any,
        r_LP_P: np.ndarray,
        min_elev: float,
        type: str,
        start_time: float = 0.0,
    ) -> None:
        """Add a location to be included in opportunity calculations.

        .. warning::
            The added location will only be considered in future calls to
            :class:`~AccessSatellite.calculate_additional_windows`; opportunities are not
            computed retroactively.

        Args:
            object: Object for with to compute opportunities.
            r_LP_P: [m] Objects planet-fixed location.
            min_elev: [rad] Minimum elevation angle for access.
            type: Category of opportunity target provides.
            start_time: [s] Time at which to start calculating opportunities for this location.
        """
        location_dict = dict(r_LP_P=r_LP_P, min_elev=min_elev, type=type)
        location_dict[type] = object  # For backwards compatibility, prefer "object" key
        location_dict["object"] = object
        location_dict["start_time"] = start_time
        self.locations_for_access_checking.append(location_dict)

    def reset_post_sim_init(self) -> None:
        """Handle initial window calculations for new simulation.

        :meta private:
        """
        super().reset_post_sim_init()
        if self.initial_generation_duration is None:
            if self.simulator.time_limit == float("inf"):
                self.initial_generation_duration = 0
            else:
                self.initial_generation_duration = self.simulator.time_limit
        self.calculate_additional_windows(self.initial_generation_duration)

    def calculate_additional_windows(self, duration: float) -> None:
        """Use a multiroot finding method to evaluate imaging windows for each location.

        Args:
            duration: Time to calculate windows from end of previous window.
        """
        if duration <= 0:
            return

        calculation_start = self.window_calculation_time
        calculation_end = self.window_calculation_time + max(
            duration, self.trajectory.dt * 2, self.generation_duration
        )
        calculation_end = self.generation_duration * np.ceil(
            calculation_end / self.generation_duration
        )

        self.logger.info(
            "Finding opportunity windows from "
            f"{calculation_start:.2f} to "
            f"{calculation_end:.2f} seconds"
        )

        # Get discrete times and positions for next trajectory segment
        self.trajectory.extend_to(calculation_end)
        r_BP_P_interp = self.trajectory.r_BP_P
        window_calc_span = np.logical_and(
            r_BP_P_interp.x >= calculation_start - 1e-9,
            r_BP_P_interp.x <= calculation_end + 1e-9,
        )  # Account for floating point error in window_calculation_time
        times = r_BP_P_interp.x[window_calc_span]
        positions = r_BP_P_interp.y[window_calc_span]

        r_max = np.max(np.linalg.norm(positions, axis=-1))
        access_dist_thresh_multiplier = 1.1
        for location in self.locations_for_access_checking:
            start_idx = max(
                np.searchsorted(times, location["start_time"], side="right") - 1, 0
            )
            times_loc = times[start_idx:]
            positions_loc = positions[start_idx:]

            alt_est = r_max - np.linalg.norm(location["r_LP_P"])
            access_dist_threshold = (
                access_dist_thresh_multiplier * alt_est / np.sin(location["min_elev"])
            )
            candidate_windows = self._find_candidate_windows(
                location["r_LP_P"], times_loc, positions_loc, access_dist_threshold
            )

            for candidate_window in candidate_windows:
                roots = self._find_elevation_roots(
                    r_BP_P_interp,
                    location["r_LP_P"],
                    location["min_elev"],
                    candidate_window,
                )
                new_windows = self._refine_window(
                    roots, candidate_window, (times_loc[0], times_loc[-1])
                )
                for new_window in new_windows:
                    self._add_window(
                        location["object"],
                        new_window,
                        type=location["type"],
                        r_LP_P=location["r_LP_P"],
                        merge_time=times_loc[0],
                    )

        self.window_calculation_time = calculation_end

    @staticmethod
    def _find_elevation_roots(
        position_interp,
        location: np.ndarray,
        min_elev: float,
        window: tuple[float, float],
        min_duration: float = 0.1,
    ):
        """Find times where the elevation is equal to the minimum elevation.

        Finds exact times where the satellite's elevation relative to a target is
        equal to the minimum elevation.
        """

        def root_fn(t):
            return -(elevation(position_interp(t), location) - min_elev)

        elev_0, elev_1 = root_fn(window[0]), root_fn(window[1])

        if elev_0 < 0 and elev_1 < 0:
            logger.warning(
                "initial_generation_duration is shorter than the maximum window length; some windows may be neglected."
            )
            return []
        elif elev_0 < 0 or elev_1 < 0:
            return [root_scalar(root_fn, bracket=window).root]
        else:
            res = minimize_scalar(root_fn, bracket=window, tol=1e-4)
            if res.fun < 0:
                window_mid = res.x
                r_open = root_scalar(root_fn, bracket=(window[0], window_mid)).root
                r_close = root_scalar(root_fn, bracket=(window_mid, window[1])).root
                if r_close - r_open > min_duration:
                    return [r_open, r_close]

        return []

    @staticmethod
    def _find_candidate_windows(
        location: np.ndarray, times: np.ndarray, positions: np.ndarray, threshold: float
    ) -> list[tuple[float, float]]:
        """Find `times` where a window is plausible.

        i.e. where a `positions` point is within `threshold` of `location`. Too big of
        a dt in times may miss windows or produce bad results.
        """
        close_times = np.linalg.norm(positions - location, axis=1) < threshold
        close_indices = np.where(close_times)[0]
        groups = np.split(close_indices, np.where(np.diff(close_indices) != 1)[0] + 1)
        groups = [group for group in groups if len(group) > 0]
        candidate_windows = []
        for group in groups:
            t_start = times[max(0, group[0] - 1)]
            t_end = times[min(len(times) - 1, group[-1] + 1)]
            candidate_windows.append((t_start, t_end))
        return candidate_windows

    @staticmethod
    def _refine_window(
        endpoints: Iterable,
        candidate_window: tuple[float, float],
        computation_window: tuple[float, float],
    ) -> list[tuple[float, float]]:
        """Detect if an exact window has been truncated by a coarse window."""
        endpoints = list(endpoints)

        # Filter endpoints that are too close
        for i, endpoint in enumerate(endpoints[0:-1]):
            if abs(endpoint - endpoints[i + 1]) < 1e-6:
                endpoints[i] = None
        endpoints = [endpoint for endpoint in endpoints if endpoint is not None]

        # Find pairs
        if len(endpoints) % 2 == 1:
            if candidate_window[0] == computation_window[0]:
                endpoints.insert(0, computation_window[0])
            elif candidate_window[-1] == computation_window[-1]:
                endpoints.append(computation_window[-1])
            else:
                return []  # Temporary fix for rare issue.

        new_windows = []
        for t1, t2 in zip(endpoints[0::2], endpoints[1::2]):
            new_windows.append((t1, t2))

        return new_windows

    def _add_window(
        self,
        object: Any,
        new_window: tuple[float, float],
        type: str,
        r_LP_P: np.ndarray,
        merge_time: Optional[float] = None,
    ):
        """Add an opportunity window.

        Args:
            object: Object to add window for
            new_window: New window for target
            type: Type of window being added
            r_LP_P: Planet-fixed location of object
            merge_time: Time at which merges with existing windows will occur. If None,
                check all windows for merges.
        """
        if new_window[0] == merge_time or merge_time is None:
            for opportunity in self.opportunities:
                if (
                    opportunity["type"] == type
                    and opportunity["object"] == object
                    and opportunity["window"][1] == new_window[0]
                ):
                    opportunity["window"] = (opportunity["window"][0], new_window[1])
                    return
        bisect.insort(
            self.opportunities,
            {"object": object, "window": new_window, "type": type, "r_LP_P": r_LP_P},
            key=lambda x: x["window"][1],
        )

    @property
    def upcoming_opportunities(self) -> list[dict]:
        """Ordered list of opportunities that have not yet closed."""
        start = bisect.bisect_left(
            self.opportunities,
            self.simulator.sim_time + 1e-12,
            key=lambda x: x["window"][1],
        )
        upcoming = self.opportunities[start:]
        return upcoming

    def opportunities_dict(
        self,
        types: Optional[Union[str, list[str]]] = None,
        filter: Union[Optional[Callable], list] = None,
    ) -> dict[Any, list[tuple[float, float]]]:
        """Make dictionary of opportunities that maps objects to lists of windows.

        Args:
            types: Types of opportunities to include. If None, include all types.
            filter: Function that takes an opportunity dictionary and returns a boolean
                if the opportunity should be included in the output.
        """
        if isinstance(types, str):
            types = [types]

        if isinstance(filter, list):
            filter_list = filter
            filter = lambda opportunity: opportunity["object"] not in filter_list

        if filter is None:
            filter = self.default_access_filter

        windows = {}
        for opportunity in self.opportunities:
            type = opportunity["type"]
            if (types is None or type in types) and filter(opportunity):
                if opportunity["object"] not in windows:
                    windows[opportunity["object"]] = []
                windows[opportunity["object"]].append(opportunity["window"])
        return windows

    def upcoming_opportunities_dict(
        self,
        types: Optional[Union[str, list[str]]] = None,
        filter: Union[Optional[Callable], list] = None,
    ) -> dict[Any, list[tuple[float, float]]]:
        """Get dictionary of upcoming opportunities.

        Maps objects to lists of windows that have not yet closed.

        Args:
            types: Types of opportunities to include. If None, include all types.
            filter: Function that takes an opportunity dictionary and returns a boolean
                if the opportunity should be included in the output.
        """
        if isinstance(types, str):
            types = [types]

        if isinstance(filter, list):
            filter_list = filter
            filter = lambda opportunity: opportunity["object"] not in filter_list

        if filter is None:
            filter = self.default_access_filter

        windows = {}
        for opportunity in self.upcoming_opportunities:
            type = opportunity["type"]
            if (types is None or type in types) and filter(opportunity):
                if opportunity["object"] not in windows:
                    windows[opportunity["object"]] = []
                windows[opportunity["object"]].append(opportunity["window"])
        return windows

    def next_opportunities_dict(
        self,
        types: Optional[Union[str, list[str]]] = None,
        filter: Union[Optional[Callable], list] = None,
        min_needed: Optional[int] = None,
    ) -> dict[Any, tuple[float, float]]:
        """Make dictionary of opportunities that maps objects to the next open windows.

        Args:
            types: Types of opportunities to include. If None, include all types.
            filter: Function that takes an opportunity dictionary and returns a boolean
                if the opportunity should be included in the output.
            min_needed: Minimum number of opportunities to return. If None, return all
        """
        if isinstance(types, str):
            types = [types]

        if isinstance(filter, list):
            filter_list = filter
            filter = lambda opportunity: opportunity["object"] not in filter_list

        if filter is None:
            filter = self.default_access_filter

        next_windows = {}
        total_found = 0
        for opportunity in self.upcoming_opportunities:
            type = opportunity["type"]
            if (types is None or type in types) and filter(opportunity):
                if opportunity["object"] not in next_windows:
                    next_windows[opportunity["object"]] = opportunity["window"]
                    total_found += 1
            if min_needed is not None and total_found >= min_needed:
                break
        return next_windows

    def find_next_opportunities(
        self,
        n: int,
        pad: bool = True,
        max_lookahead: int = 100,
        types: Optional[Union[str, list[str]]] = None,
        filter: Union[Optional[Callable], list] = None,
    ) -> list[dict]:
        """Find the n nearest opportunities, sorted by window close time.

        Args:
            n: Number of opportunities to attempt to include.
            pad: If true, duplicates the last target if the number of opportunities
                found is less than n.
            max_lookahead: Maximum times to call calculate_additional_windows.
            types: Types of opportunities to include. If None, include all types.
            filter: Function that takes an opportunity dictionary and returns a boolean
                if the opportunity should be included in the output.

        Returns:
            ``n`` nearest opportunities, ordered
        """
        if isinstance(types, str):
            types = [types]

        if isinstance(filter, list):
            filter_list = filter
            filter = lambda opportunity: opportunity["object"] not in filter_list

        if filter is None:
            filter = self.default_access_filter

        if n == 0:
            return []

        for _ in range(max_lookahead):
            upcoming_opportunities = self.upcoming_opportunities
            next_opportunities = []
            for opportunity in upcoming_opportunities:
                type = opportunity["type"]
                if (types is None or type in types) and filter(opportunity):
                    next_opportunities.append(opportunity)

                if len(next_opportunities) >= n:
                    return next_opportunities
            if (
                self.window_calculation_time
                >= self.initial_generation_duration
                + self.max_generation_duration_beyond_initial
            ):
                break
            self.calculate_additional_windows(self.generation_duration)
        if pad and len(next_opportunities) >= 1:
            self.logger.info(
                f"Only {len(next_opportunities)} opportunities found, padding to {n}."
            )
            next_opportunities += [next_opportunities[-1]] * (
                n - len(next_opportunities)
            )
        else:
            raise RuntimeError(
                "No opportunities found! Use add_location_for_access_checking to add locations."
            )
        return next_opportunities

    def get_access_filter(self):
        """Deprecated function.

        :meta private:
        """
        raise DeprecationWarning(
            "get_access_filter is deprecated. Use add_access_filter and default_access_filter instead."
        )

    def add_access_filter(
        self,
        access_filter_fn: Callable,
        types: Optional[Union[str, list[str]]] = None,
        prepend: bool = False,
    ):
        """Add an access filter function to the list of access filters.

        Calls to :class:`~AccessSatellite.opportunities_dict`, :class:`~AccessSatellite.find_next_opportunities`,
        and similar functions will use the boolean AND of all access filter functions,
        unless otherwise specified.

        Access filters are used by various other aspects of the environment to limit
        which opportunities are considered based on the satellite's local knowledge of
        the environment.
        """
        if types is not None:
            if isinstance(types, str):
                types = [types]

            def access_filter_type_restricted(opportunity):
                return opportunity["type"] not in types or access_filter_fn(opportunity)

            to_add = access_filter_type_restricted
        else:
            to_add = access_filter_fn

        if prepend:
            self.access_filter_functions.insert(0, to_add)
        else:
            self.access_filter_functions.append(to_add)

    @property
    def default_access_filter(self):
        """Generate a default access filter function that combines all access filters.

        :meta private:
        """

        def access_filter(opportunity):
            for access_filter_fn in self.access_filter_functions:
                if not access_filter_fn(opportunity):
                    return False
            else:
                return True

        return access_filter


class ImagingSatellite(AccessSatellite):
    """Satellite with agile imaging capabilities."""

    buffer_name = "image_buffer"

    dyn_type = dyn.ImagingDynModel
    fsw_type = fsw.ImagingFSWModel

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Satellite with agile imaging capabilities.

        Stop the simulation when a target is imaged or missed so that time is not wasted
        on an inaccessible or already imaged target.
        """
        super().__init__(*args, **kwargs)
        self.fsw: ImagingSatellite.fsw_type
        self.dynamics: ImagingSatellite.dyn_type
        self.data_store: "UniqueImageStore"
        self.target_types = "target"
        self.latest_target = None

    @property
    def known_targets(self) -> list["Target"]:
        """List of known targets."""
        try:
            return self.data_store.data.known
        except AttributeError:
            return []

    def reset_overwrite_previous(self) -> None:
        """Overwrite statistics about previous episode."""
        super().reset_overwrite_previous()
        self._image_event_name = None
        self.imaged = 0
        self.missed = 0

    def reset_pre_sim_init(self) -> None:
        """Set the buffer parameters based on computed windows.

        :meta private:
        """
        super().reset_pre_sim_init()
        self.sat_args["bufferNames"] = [self.buffer_name]
        self.sat_args["transmitterNumBuffers"] = len(self.sat_args["bufferNames"])

    def _update_image_event(self, target: "Target") -> None:
        """Create a simulator event that terminates on imaging.

        Causes the simulation to stop when a target is imaged.

        Args:
            target: Target expected to be imaged
        """
        self._disable_image_event()

        self._image_event_name = valid_func_name(f"image_{self.name}_{target.id}")
        if self._image_event_name not in self.simulator.eventMap.keys():
            current_data_level = (
                self.dynamics.storageUnit.storageUnitDataOutMsg.read().storedData[0]
            )

            def side_effect(sim):
                self.logger.info(f"imaged {target}")
                self.imaged += 1
                self.requires_retasking = True
                self.remove_imaging_line()

            self.simulator.createNewEvent(
                self._image_event_name,
                macros.sec2nano(self.fsw.fsw_rate),
                True,
                conditionFunction=lambda sim: self.dynamics.storageUnit.storageUnitDataOutMsg.read().storedData[
                    0
                ]
                > current_data_level,
                actionFunction=side_effect,
                terminal=self.variable_interval,
            )
        else:
            self.simulator.eventMap[self._image_event_name].eventActive = True

    def _disable_image_event(self) -> None:
        """Turn off simulator termination due to this satellite's imaging checker."""
        if (
            self._image_event_name is not None
            and self._image_event_name in self.simulator.eventMap
        ):
            self.simulator.delete_event(self._image_event_name)

    def parse_target_selection(self, target_query: Union[int, Target, str]):
        """Identify a target from a query.

        Parses an upcoming target index, Target object, or target id.

        Args:
            target_query: Target upcoming index, object, or id.
        """
        if np.issubdtype(type(target_query), np.integer):
            target = self.find_next_opportunities(
                n=target_query + 1, types=self.target_types
            )[-1]["object"]
        elif isinstance(target_query, Target):
            target = target_query
        elif isinstance(target_query, str):
            try:
                target = [
                    target for target in self.known_targets if target.id == target_query
                ][0]
            except IndexError:
                raise ValueError(f"Target {target_query} not a known target!")
        else:
            raise TypeError(f"Invalid target_query! Cannot be a {type(target_query)}!")

        return target

    def enable_target_window(
        self, target: "Target", max_duration: Optional[float] = None
    ):
        """Enable a timed opportunity close event and a successfully imaged event.

        Args:
            target: Target to terminate the step on imaging or when out of range.
            max_duration: [s] Maximum duration to wait for imaging. If None, use the default
                behavior.
        """
        self._update_image_event(target)
        for opportunity in self.upcoming_opportunities:
            if opportunity["object"] == target:
                next_window = opportunity["window"]
                break
        self.logger.info(
            f"{target} window enabled: {next_window[0]:.1f} to {next_window[1]:.1f}"
        )
        if max_duration is None:
            max_duration = 1e9

        def side_effect(sim):
            if np.isclose(sim.sim_time, next_window[1], atol=1e-9):
                self.missed += 1
            self.remove_imaging_line()

        self.update_timed_terminal_event(
            min(next_window[1], self.simulator.sim_time + max_duration),
            info=f"for {target} window",
            extra_actions=side_effect,
        )

    def task_target_for_imaging(
        self, target: "Target", max_duration: Optional[float] = None
    ):
        """Task the satellite to image a target.

        Args:
            target: Selected target
            max_duration: [s] Maximum duration to wait for imaging. If None, wait until
                the end of the target's access window.
        """
        msg = f"{target} tasked for imaging"
        self.logger.info(msg)
        self.fsw.action_image(target.r_LP_P, self.buffer_name)
        self.enable_target_window(target, max_duration=max_duration)
        self.draw_imaging_line(target)
        self.latest_target = target

    @vizard.visualize
    def draw_imaging_line(
        self, target: "Target", vizSupport=None, vizInstance=None
    ) -> None:
        """Draw a line from the satellite to the target in vizard."""
        if not hasattr(self, "target_line"):
            vizSupport.createTargetLine(
                vizInstance,
                fromBodyName=self.name,
                toBodyName=target.name,
                lineColor=self.vizard_color,
            )
            self.target_line = vizSupport.targetLineList[-1]
        self.target_line.toBodyName = target.name
        vizSupport.updateTargetLineList(vizInstance)

    @vizard.visualize
    def remove_imaging_line(self, vizSupport=None, vizInstance=None):
        """Remove the imaging line from Vizard."""
        if hasattr(self, "target_line"):
            self.target_line.toBodyName = self.name
            vizSupport.updateTargetLineList(vizInstance)



"""FSW models for orbital maneuvers."""

import numpy as np

from bsk_rl.sim.fsw import BasicFSWModel, action
from bsk_rl.utils.functional import aliveness_checker, default_args


class MagicOrbitalManeuverFSWModel(BasicFSWModel):
    """Model that allows for instantaneous Delta V maneuvers."""

    def __init__(self, *args, **kwargs) -> None:
        """Model that allows for instantaneous Delta V maneuvers."""
        super().__init__(*args, **kwargs)
        self.setup_fuel(**kwargs)
        self.thrust_count = 0

    @property
    def dv_available(self):
        """Delta-V available for the satellite."""
        return self._dv_available

    @aliveness_checker
    def fuel_remaining(self) -> bool:
        """Check if the satellite has fuel remaining."""
        return self.dv_available > 1e-8

    @default_args(dv_available_init=100.0)
    def setup_fuel(self, dv_available_init: float, **kwargs):
        """Set up available fuel for the satellite.

        Args:
            dv_available_init: [m/s] Initial fuel level.
            kwargs: Passed to other setup functions.
        """
        # TODO: may adjust names for consistency with modelled fuel take in future.
        self._dv_available = dv_available_init

    @action
    def action_impulsive_thrust(self, dv_N: np.ndarray) -> None:
        """Thrust relative to the inertial frame.

        Args:
            dv_N: [m/s] Inertial Delta V.
        """
        if np.linalg.norm(dv_N) > self.dv_available:
            self.satellite.logger.warning(
                f"Maneuver exceeds available Delta V ({np.linalg.norm(dv_N)}/{self.dv_available} m/s)."
            )
            dv_N = dv_N / np.linalg.norm(dv_N) * self.dv_available

        self._dv_available -= np.linalg.norm(dv_N)

        self.dynamics.scObject.dynManager.getStateObject(
            self.dynamics.scObject.hub.nameOfHubVelocity
        ).setState(list(np.array(self.dynamics.v_BN_N) + np.array(dv_N)))

        self.thrust_count += 1



import gymnasium as gym
import numpy as np
from pytest import approx

from bsk_rl import act, data, obs, sats
from bsk_rl.scene import UniformTargets
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import random_orbit


##############################
# Composed Observation Tests #
##############################
class TestComposedState:
    class ComposedPropSat(sats.ImagingSatellite):
        dyn_type = dyn.ImagingDynModel
        fsw_type = fsw.ImagingFSWModel
        observation_spec = [
            obs.Time(),
            obs.SatProperties(
                dict(prop="r_BN_N", module="dynamics"),
                dict(prop="r_BN_N", norm=7000.0 * 1e3),
            ),
            obs.OpportunityProperties(dict(prop="priority"), n_ahead_observe=2),
            obs.Eclipse(),
        ]
        action_spec = [act.Drift()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ComposedPropSat(
            "Explorer 1",
            sat_args=ComposedPropSat.default_sat_args(oe=random_orbit),
        ),
        scenario=UniformTargets(n_targets=1000),
        rewarder=data.UniqueImageReward(),
        sim_rate=1.0,
        max_step_duration=10.0,
        time_limit=100.0,
        disable_env_checker=True,
    )

    def test_normd_property_state(self):
        observation, info = self.env.reset()
        assert observation[0] == 0.0  # Timed observation should be first
        self.env.unwrapped.satellite.observation_builder.obs_type = dict
        eclipse = self.env.unwrapped.satellite.get_obs()["eclipse"]
        assert observation[-2] == eclipse[0]  # Eclipse should be last
        assert observation[-1] == eclipse[1]  # Eclipse should be last


################################
# Individual Observation Tests #
################################


class TestSatProperties:
    class SatPropertiesSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.BasicFSWModel
        observation_spec = [
            obs.SatProperties(
                dict(prop="r_BN_N", module="dynamics"),
                dict(prop="r_BN_N", norm=7000.0 * 1e3),
                dict(prop="inclination", module="dynamics", norm=np.pi / 180),
            ),
        ]
        action_spec = [act.Drift()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=SatPropertiesSat(
            "Sputnik",
            sat_args=SatPropertiesSat.default_sat_args(oe=random_orbit(a=7000, i=45)),
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        max_step_duration=10.0,
        time_limit=100.0,
        disable_env_checker=True,
    )

    def test_normd_property_state(self):
        observation, info = self.env.reset()
        assert np.linalg.norm(observation[0:3]) == approx(7000.0 * 1e3)
        assert np.linalg.norm(observation[3:6]) == approx(1.0)
        assert observation[6] == approx(45)
        observation, reward, terminated, truncated, info = self.env.step(0)
        assert observation[6] == approx(45)


class TestTime:
    class TimedSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.BasicFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Drift()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=TimedSat(
            "Voyager",
            sat_args=TimedSat.default_sat_args(oe=random_orbit()),
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        max_step_duration=10.0,
        time_limit=100.0,
        disable_env_checker=True,
    )

    def test_normd_property_state(self):
        observation, info = self.env.reset()
        assert observation[0] == 0.0
        observation, reward, terminated, truncated, info = self.env.step(0)
        assert observation[0] == 0.1


class TestOpportunityProperties:
    class TargetSat(sats.ImagingSatellite):
        dyn_type = dyn.ImagingDynModel
        fsw_type = fsw.ImagingFSWModel
        observation_spec = [
            obs.OpportunityProperties(dict(prop="priority"), n_ahead_observe=2)
        ]
        action_spec = [act.Drift()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=TargetSat(
            "Bullseye",
            obs_type=dict,
            sat_args=TargetSat.default_sat_args(oe=random_orbit()),
        ),
        scenario=UniformTargets(n_targets=100),
        rewarder=data.UniqueImageReward(),
        sim_rate=1.0,
        max_step_duration=10.0,
        time_limit=100.0,
        disable_env_checker=True,
    )

    def test_target_state(self):
        observation, info = self.env.reset()
        assert "target_1" in observation["target"]
        assert "priority" in observation["target"]["target_1"]


class TestEclipse:
    class EclipseSat(sats.Satellite):
        dyn_type = dyn.BasicDynamicsModel
        fsw_type = fsw.BasicFSWModel
        observation_spec = [obs.Eclipse()]
        action_spec = [act.Drift()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=EclipseSat(
            "PinkFloyd",
            obs_type=list,
            sat_args=EclipseSat.default_sat_args(oe=random_orbit()),
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        max_step_duration=5700.0,
        time_limit=5800.0,
        disable_env_checker=True,
    )

    def test_eclipse_state(self):
        observation1, info = self.env.reset()
        observation2, reward, terminated, truncated, info = self.env.step(0)
        assert (observation2[0] - observation1[0]) < 0.05
        assert (observation2[1] - observation1[1]) < 0.05


class TestGroundStationProperties:
    class GroundSat(sats.AccessSatellite):
        dyn_type = dyn.GroundStationDynModel
        fsw_type = fsw.ImagingFSWModel
        observation_spec = [
            obs.OpportunityProperties(
                dict(
                    prop="opportunity_open",
                ),
                n_ahead_observe=2,
                type="ground_station",
            ),
        ]
        action_spec = [act.Drift()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=GroundSat(
            "Satellite",
            obs_type=list,
            sat_args=GroundSat.default_sat_args(oe=random_orbit()),
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=data.NoReward(),
        sim_rate=1.0,
        max_step_duration=5700.0,
        time_limit=5700.0,
        disable_env_checker=True,
    )

    def test_ground_station_state(self):
        observation, info = self.env.reset()
        assert sum(observation) > 0  # Check that there are downlink opportunities


class TestResourceRewardWeight:
    class ResourceSat(sats.ImagingSatellite):
        dyn_type = dyn.ImagingDynModel
        fsw_type = fsw.ImagingFSWModel
        observation_spec = [obs.ResourceRewardWeight()]
        action_spec = [act.Drift()]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ResourceSat(
            "ResourceSat",
            obs_type=dict,
        ),
        scenario=UniformTargets(n_targets=0),
        rewarder=(
            data.ResourceReward(
                resource_fn=lambda sat: sat.simulator.sim_time, reward_weight=0.1
            ),
            data.ResourceReward(
                resource_fn=lambda sat: sat.simulator.sim_time, reward_weight=1.0
            ),
        ),
        sim_rate=1.0,
        max_step_duration=10.0,
        disable_env_checker=True,
    )

    def test_resource_reward_weight(self):
        observation, info = self.env.reset()
        observation, reward, terminated, truncated, info = self.env.step(0)
        assert reward == 1.0 + 10.0
        assert (observation["resource_reward_weight"] == [0.1, 1.0]).all()



"""Basilisk world models are given in ``bsk_rl.sim.world``.

In most cases, the user does not need to specify the world model, as it is inferred from
the requirements of the :class:`~bsk_rl.sim.fsw.FSWModel`. However, the user can specify
the world model in the :class:`~bsk_rl.GeneralSatelliteTasking` constructor if desired.

Customization of the world model parameters is via the ``world_args`` parameter in the
:class:`~bsk_rl.GeneralSatelliteTasking`. As with ``sat_args``, these parameters are
passed as a dictionary of key-value or key-function pairs, with the latter called to
generate the value each time the simulation is reset.

.. code-block:: python

    world_args = dict(
        utc_init="2018 SEP 29 21:00:00.000 (UTC)",  # set the epoch
        scaleHeight=np.random.uniform(7e3, 9e3),  # randomize the atmosphere
    )

In general, ``world_args`` parameter names match those used in Basilisk. See the setup
functions for short descriptions of what parameters do and the Basilisk documentation
for more detail on their exact model effects.

"""

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Union
from weakref import proxy

import numpy as np
from Basilisk import __path__
from Basilisk.simulation import (
    eclipse,
    ephemerisConverter,
    exponentialAtmosphere,
    groundLocation,
)
from Basilisk.utilities import macros as mc
from Basilisk.utilities import orbitalMotion, simIncludeGravBody

from bsk_rl.utils.functional import collect_default_args, default_args
from bsk_rl.utils.orbital import random_epoch

if TYPE_CHECKING:  # pragma: no cover
    from bsk_rl.sim import Simulator

logger = logging.getLogger(__name__)

bsk_path = __path__[0]


class WorldModel(ABC):
    """Abstract Basilisk world model."""

    @classmethod
    def default_world_args(cls, **kwargs) -> dict[str, Any]:
        """Compile default arguments for the world model.

        Args:
            **kwargs: Arguments to override in the default arguments.

        Returns:
            Dictionary of arguments for simulation models.
        """
        defaults = collect_default_args(cls)
        for k, v in kwargs.items():
            if k not in defaults:
                raise KeyError(f"{k} not a valid key for world_args")
            defaults[k] = v
        return defaults

    def __init__(
        self,
        simulator: "Simulator",
        world_rate: float,
        priority: int = 300,
        **kwargs,
    ) -> None:
        """Abstract Basilisk world model.

        One WorldModel is instantiated for the environment each time a new simulator
        is created.

        Args:
            simulator: Simulator using this model.
            world_rate: Rate of world simulation [s]
            priority: Model priority.
            kwargs: Passed through to setup functions.
        """
        self.simulator: Simulator = proxy(simulator)

        world_proc_name = "WorldProcess"
        world_proc = self.simulator.CreateNewProcess(world_proc_name, priority)

        # Define process name, task name and task time-step
        self.world_task_name = "WorldTask"
        world_proc.addTask(
            self.simulator.CreateNewTask(self.world_task_name, mc.sec2nano(world_rate))
        )

        self._setup_world_objects(**kwargs)

    def __del__(self):
        """Log when world is deleted."""
        logger.debug("Basilisk world deleted")

    @abstractmethod  # pragma: no cover
    def _setup_world_objects(self, **kwargs) -> None:
        """Caller for all world objects."""
        pass


class BasicWorldModel(WorldModel):
    """Basic world with minimum necessary Basilisk world components."""

    def __init__(self, *args, **kwargs) -> None:
        """Basic world with minimum necessary Basilisk world components.

        This model includes ephemeris and SPICE-based Earth gravity and dynamics models,
        an exponential atmosphere model, and an eclipse model.

        Args:
            *args: Passed to superclass.
            **kwargs: Passed to superclass.
        """
        super().__init__(*args, **kwargs)

    @property
    def PN(self):
        """Planet relative to inertial frame rotation matrix."""
        return np.array(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.body_index]
            .read()
            .J20002Pfix
        ).reshape((3, 3))

    @property
    def omega_PN_N(self):
        """Planet angular velocity in inertial frame [rad/s]."""
        PNdot = np.array(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.body_index]
            .read()
            .J20002Pfix_dot
        ).reshape((3, 3))
        skew_PN_N = -np.matmul(np.transpose(self.PN), PNdot)
        return np.array([skew_PN_N[2, 1], skew_PN_N[0, 2], skew_PN_N[1, 0]])

    def _setup_world_objects(self, **kwargs) -> None:
        self.setup_gravity_bodies(**kwargs)
        self.setup_ephem_object(**kwargs)
        self.setup_atmosphere_density_model(**kwargs)
        self.setup_eclipse_object(**kwargs)

    @default_args(utc_init=random_epoch)
    def setup_gravity_bodies(
        self, utc_init: str, priority: int = 1100, **kwargs
    ) -> None:
        """Specify gravitational models to use in the simulation.

        Args:
            utc_init: UTC datetime string, in the format ``YYYY MMM DD hh:mm:ss.sss (UTC)``
            priority: Model priority.
            **kwargs: Passed to other setup functions.
        """
        self.gravFactory = simIncludeGravBody.gravBodyFactory()
        self.gravFactory.createSun()
        self.planet = self.gravFactory.createEarth()
        self.sun_index = 0
        self.body_index = 1

        self.planet.isCentralBody = (
            True  # ensure this is the central gravitational body
        )
        self.planet.useSphericalHarmonicsGravityModel(
            bsk_path + "/supportData/LocalGravData/GGM03S.txt", 10
        )

        # setup Spice interface for some solar system bodies
        timeInitString = utc_init
        self.gravFactory.createSpiceInterface(
            bsk_path + "/supportData/EphemerisData/", timeInitString, epochInMsg=True
        )
        self.gravFactory.spiceObject.zeroBase = "earth"

        self.simulator.AddModelToTask(
            self.world_task_name, self.gravFactory.spiceObject, ModelPriority=priority
        )

    def setup_ephem_object(self, priority: int = 988, **kwargs) -> None:
        """Set up the ephemeris object to use with the SPICE library.

        Args:
            priority: Model priority.
            **kwargs: Passed to other setup functions.
        """
        self.ephemConverter = ephemerisConverter.EphemerisConverter()
        self.ephemConverter.ModelTag = "ephemConverter"
        self.ephemConverter.addSpiceInputMsg(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.sun_index]
        )
        self.ephemConverter.addSpiceInputMsg(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.body_index]
        )
        self.simulator.AddModelToTask(
            self.world_task_name, self.ephemConverter, ModelPriority=priority
        )

    @default_args(
        planetRadius=orbitalMotion.REQ_EARTH * 1e3,
        baseDensity=1.22,
        scaleHeight=8e3,
    )
    def setup_atmosphere_density_model(
        self,
        planetRadius: float,
        baseDensity: float,
        scaleHeight: float,
        priority: int = 1000,
        **kwargs,
    ) -> None:
        """Set up the exponential gravity model.

        Args:
            planetRadius: [m] Planet ground radius.
            baseDensity: [kg/m^3] Exponential model parameter.
            scaleHeight: [m] Exponential model parameter.
            priority: Model priority.
            **kwargs: Passed to other setup functions.
        """
        self.densityModel = exponentialAtmosphere.ExponentialAtmosphere()
        self.densityModel.ModelTag = "expDensity"
        self.densityModel.planetRadius = planetRadius
        self.densityModel.baseDensity = baseDensity
        self.densityModel.scaleHeight = scaleHeight
        self.densityModel.planetPosInMsg.subscribeTo(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.body_index]
        )
        self.simulator.AddModelToTask(
            self.world_task_name, self.densityModel, ModelPriority=priority
        )

    def setup_eclipse_object(self, priority: int = 988, **kwargs) -> None:
        """Set up the celestial object that is causing an eclipse message.

        Args:
            priority: Model priority.
            kwargs: Ignored
        """
        self.eclipseObject = eclipse.Eclipse()
        self.eclipseObject.addPlanetToModel(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.body_index]
        )
        self.eclipseObject.sunInMsg.subscribeTo(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.sun_index]
        )
        self.simulator.AddModelToTask(
            self.world_task_name, self.eclipseObject, ModelPriority=priority
        )

    def __del__(self) -> None:
        """Log when world is deleted and unload SPICE."""
        super().__del__()
        try:
            self.gravFactory.unloadSpiceKernels()
        except AttributeError:
            pass


class GroundStationWorldModel(BasicWorldModel):
    """Model that includes downlink ground stations."""

    def __init__(self, *args, **kwargs) -> None:
        """Model that includes downlink ground stations.

        This model includes the basic world components, as well as ground stations for
        downlinking data.

        Args:
            *args: Passed to superclass.
            **kwargs: Passed to superclass.
        """
        super().__init__(*args, **kwargs)

    def _setup_world_objects(self, **kwargs) -> None:
        super()._setup_world_objects(**kwargs)
        self.setup_ground_locations(**kwargs)

    @default_args(
        groundStationsData=[
            dict(name="Boulder", lat=40.009971, long=-105.243895, elev=1624),
            dict(name="Merritt", lat=28.3181, long=-80.6660, elev=0.9144),
            dict(name="Singapore", lat=1.3521, long=103.8198, elev=15),
            dict(name="Weilheim", lat=47.8407, long=11.1421, elev=563),
            dict(name="Santiago", lat=-33.4489, long=-70.6693, elev=570),
            dict(name="Dongara", lat=-29.2452, long=114.9326, elev=34),
            dict(name="Hawaii", lat=19.8968, long=-155.5828, elev=9),
        ],
        groundLocationPlanetRadius=orbitalMotion.REQ_EARTH * 1e3,
        gsMinimumElevation=np.radians(10.0),
        gsMaximumRange=-1,
    )
    def setup_ground_locations(
        self,
        groundStationsData: list[dict[str, Union[str, float]]],
        groundLocationPlanetRadius: float,
        gsMinimumElevation: float,
        gsMaximumRange: float,
        priority: int = 1399,
        **kwargs,
    ) -> None:
        """Specify the ground locations of interest.

        Args:
            groundStationsData: List of dictionaries of ground station data. Each dictionary
                must include keys for ``lat`` and ``long`` [deg], and may include
                ``elev`` [m], ``name``. For example:

                .. code-block:: python

                    groundStationsData=[
                        dict(name="Boulder", lat=40.009971, long=-105.243895, elev=1624),
                        dict(lat=28.3181, long=-80.6660),
                    ]

                ``groundLocationPlanetRadius``, ``gsMinimumElevation``, and ``gsMaximumRange``
                may also be specified in the dictionary to override the global values
                for those parameters for a specific ground station.

            groundLocationPlanetRadius: [m] Radius of ground locations from center of
                planet.
            gsMinimumElevation: [rad] Minimum elevation angle from station to satellite
                to be able to downlink data.
            gsMaximumRange: [m] Maximum range from station to satellite when
                downlinking. Set to ``-1`` to disable.
            priority: Model priority.
            kwargs: Passed to other setup functions.
        """
        self.groundStations = []
        self.groundLocationPlanetRadius = groundLocationPlanetRadius
        self.gsMinimumElevation = gsMinimumElevation
        self.gsMaximumRange = gsMaximumRange
        for i, groundStationData in enumerate(groundStationsData):
            self._create_ground_station(**groundStationData, priority=priority - i)

    def _create_ground_station(
        self,
        lat: float,
        long: float,
        elev: float = 0,
        name: Optional[str] = None,
        groundLocationPlanetRadius: Optional[float] = None,
        gsMinimumElevation: Optional[float] = None,
        gsMaximumRange: Optional[float] = None,
        priority: int = 1399,
    ) -> None:
        """Add a ground station with given parameters.

        Args:
            lat: [deg] Latitude.
            long: [deg] Longitude.
            elev: [m] Elevation.
            name: Ground station identifier.
            groundLocationPlanetRadius: [m] Radius of planet.
            gsMinimumElevation: [rad] Minimum elevation angle to downlink to ground station.
            gsMaximumRange: [m] Maximum range to downlink to ground station. Set to ``-1`` for infinity.
            priority: Model priority.
        """
        if name is None:
            name = str(len(self.groundStations))

        groundStation = groundLocation.GroundLocation()
        groundStation.ModelTag = "GroundStation" + name
        if groundLocationPlanetRadius:
            groundStation.planetRadius = groundLocationPlanetRadius
        else:
            groundStation.planetRadius = self.groundLocationPlanetRadius
        groundStation.specifyLocation(np.radians(lat), np.radians(long), elev)
        groundStation.planetInMsg.subscribeTo(
            self.gravFactory.spiceObject.planetStateOutMsgs[self.body_index]
        )
        if gsMinimumElevation:
            groundStation.minimumElevation = gsMinimumElevation
        else:
            groundStation.minimumElevation = self.gsMinimumElevation
        if gsMaximumRange:
            groundStation.maximumRange = gsMaximumRange
        else:
            groundStation.maximumRange = self.gsMaximumRange
        self.groundStations.append(groundStation)

        self.simulator.AddModelToTask(
            self.world_task_name, groundStation, ModelPriority=priority
        )


__doc_title__ = "World Sims"
__all__ = ["WorldModel", "BasicWorldModel", "GroundStationWorldModel"]



from warnings import warn

import gymnasium as gym
import numpy as np
import pytest
from pettingzoo.test.parallel_test import parallel_api_test

from bsk_rl import ConstellationTasking, act, data, obs, sats, scene
from bsk_rl.utils.orbital import random_orbit


class FullFeaturedSatellite(sats.ImagingSatellite):
    observation_spec = [
        obs.SatProperties(dict(prop="r_BN_P", module="dynamics", norm=6e6)),
        obs.Time(),
    ]
    action_spec = [act.Image(n_ahead_image=10)]


multi_env = gym.make(
    "GeneralSatelliteTasking-v1",
    satellites=[
        FullFeaturedSatellite(
            "Sentinel-2A",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.01,
                imageRateErrorRequirement=0.01,
            ),
        ),
        FullFeaturedSatellite(
            "Sentinel-2B",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.01,
                imageRateErrorRequirement=0.01,
            ),
        ),
    ],
    scenario=scene.UniformTargets(n_targets=1000),
    rewarder=data.UniqueImageReward(),
    sim_rate=0.5,
    max_step_duration=1e9,
    time_limit=5700.0,
    disable_env_checker=True,
)

parallel_env = ConstellationTasking(
    satellites=[
        FullFeaturedSatellite(
            "Sentinel-2A",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.01,
                imageRateErrorRequirement=0.01,
            ),
        ),
        FullFeaturedSatellite(
            "Sentinel-2B",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.01,
                imageRateErrorRequirement=0.01,
            ),
        ),
    ],
    scenario=scene.UniformTargets(n_targets=1000),
    rewarder=data.UniqueImageReward(),
    sim_rate=0.5,
    max_step_duration=1e9,
    time_limit=5700.0,
)

parallel_meta_env = ConstellationTasking(
    satellites=[
        FullFeaturedSatellite(
            "Sentinel-2A",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.01,
                imageRateErrorRequirement=0.01,
            ),
        ),
        FullFeaturedSatellite(
            "Sentinel-2B",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.01,
                imageRateErrorRequirement=0.01,
            ),
        ),
        FullFeaturedSatellite(
            "Sentinel-2C",
            sat_args=FullFeaturedSatellite.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.01,
                imageRateErrorRequirement=0.01,
            ),
        ),
    ],
    scenario=scene.UniformTargets(n_targets=1000),
    rewarder=data.UniqueImageReward(),
    sim_rate=0.5,
    max_step_duration=1e9,
    time_limit=5700.0,
    meta_agent_groupings={
        "DoubleSat": ["Sentinel-2A", "Sentinel-2C"],
        "SingleSat": ["Sentinel-2B"],
    },
)


@pytest.mark.parametrize("env", [multi_env])
def test_reproducibility(env):
    actions = [env.action_space.sample() for _ in range(1000)]
    reward_sum_1 = 0
    observation_1, info = env.reset(seed=0)
    for action in actions:
        observation, reward, terminated, truncated, info = env.step(action)
        reward_sum_1 += reward
        if truncated or terminated:
            break

    truncated = False
    terminated = False
    reward_sum_2 = 0
    observation_2, info = env.reset(seed=0)
    for o1, o2 in zip(observation_1, observation_2):
        np.testing.assert_allclose(o1, o2)

    for action in actions:
        observation, reward, terminated, truncated, info = env.step(action)
        reward_sum_2 += reward
        if truncated or terminated:
            break

    assert reward_sum_2 == reward_sum_1


@pytest.mark.repeat(5)
@pytest.mark.parametrize("env", [parallel_env, parallel_meta_env])
def test_parallel_api(env):
    with pytest.warns(UserWarning):
        # expect an erroneous warning about the info dict due to our additional info
        try:
            parallel_api_test(env)
        except AssertionError as e:
            if "agent cannot be revived once dead" in str(e):
                warn(f"'{e}' is a known issue (#59)")
            else:
                raise (e)



import gymnasium as gym

from bsk_rl import act, data, obs, sats
from bsk_rl.scene import CityTargets, UniformTargets
from bsk_rl.sim import dyn, fsw
from bsk_rl.utils.orbital import random_orbit


def make_env(scenario):
    class ImageSat(sats.ImagingSatellite):
        dyn_type = dyn.GroundStationDynModel
        fsw_type = fsw.ImagingFSWModel
        observation_spec = [obs.Time()]
        action_spec = [act.Image(n_ahead_image=10)]

    env = gym.make(
        "SatelliteTasking-v1",
        satellite=ImageSat(
            "EO-1",
            sat_args=ImageSat.default_sat_args(
                oe=random_orbit,
                imageAttErrorRequirement=0.05,
                imageRateErrorRequirement=0.05,
            ),
        ),
        scenario=scenario,
        rewarder=data.UniqueImageReward(),
        sim_rate=1.0,
        time_limit=5700.0,
        max_step_duration=1e9,
        disable_env_checker=True,
        failure_penalty=0,
    )
    return env


class TestUniformTargets:
    def test_priority_dist(self):
        env = make_env(UniformTargets(n_targets=1000, priority_distribution=lambda: 1))
        env.reset()
        for i in range(10):
            observation, reward, terminated, truncated, info = env.step(i)
            assert reward in [0.0, 1.0]


class TestCityTargets:
    def test_city_distribution(self):
        # Just a smoke test
        env = make_env(
            CityTargets(n_targets=1000, n_select_from=10000, location_offset=1e3)
        )
        env.reset()
        reward_sum = 0.0
        for i in range(10):
            observation, reward, terminated, truncated, info = env.step(i)
            reward_sum += reward
        assert reward_sum > 0



"""A collection of utilities at ``bsk_rl.utils.rllib``.

* :ref:`bsk_rl.utils.rllib.discounting` - For semi-MDP discounting with GAE.
* :ref:`bsk_rl.utils.rllib.callbacks` - For logging data at the end of each episode.

Two environments are added to the ``ray.tune.registry`` with this import. They are
``"SatelliteTasking-RLlib"`` and ``"ConstellationTasking-RLlib"``. These environments
are wrapped with the :func:`unpack_config` function to make them compatible with RLlib's
API, and they are wrapped with the :class:`EpisodeDataWrapper` to allow for data logging
at the end of each episode during training. These environments can be selected by name
when setting ``config.environment(env="SatelliteTasking-RLlib")``. Callback functions
that are arguments to :class:`EpisodeDataWrapper` can be set in the ``env_config`` dictionary.
"""

from pathlib import Path
from typing import Callable

import numpy as np
import torch
from ray.rllib.core.columns import Columns
from ray.rllib.core.rl_module.rl_module import RLModule
from ray.rllib.env.wrappers.pettingzoo_env import ParallelPettingZooEnv
from ray.rllib.utils.numpy import convert_to_numpy
from ray.rllib.utils.spaces.space_utils import unsquash_action
from ray.tune.registry import register_env

from bsk_rl import ConstellationTasking, GeneralSatelliteTasking, SatelliteTasking
from bsk_rl.utils.rllib.callbacks import EpisodeDataParallelWrapper, EpisodeDataWrapper


def load_torch_mlp_policy(
    policy_path: Path, policy_name: str = "default_agent"
) -> Callable:
    """Load a PyTorch policy from a saved model.

    Args:
        policy_path: Path to the directory containing the policy checkpoint.
        policy_name: Name of the policy to load from the checkpoint.
    """
    rl_module = RLModule.from_checkpoint(
        Path(policy_path) / "learner_group" / "learner" / "rl_module" / policy_name
    )
    cat = rl_module.config.get_catalog()
    action_dist_cls = cat.get_action_dist_cls(framework="torch")

    def policy(obs: list[float], deterministic: bool = True) -> np.ndarray:
        """Policy function that takes observations and returns actions.

        Args:
            obs: Observation vector.
            deterministic: If True, use loc for action selection; otherwise, sample from the action distribution.
        """
        obs = np.array(obs, dtype=np.float32)
        action_dist_params = rl_module.forward_inference(
            dict(obs=torch.tensor(obs)[None, :])
        )
        action_dist = action_dist_cls.from_logits(
            action_dist_params[Columns.ACTION_DIST_INPUTS]
        )

        if deterministic:
            action_squashed = action_dist.to_deterministic().sample()
        else:
            action_squashed = action_dist.sample()

        action_squashed = convert_to_numpy(action_squashed[0])
        action = unsquash_action(action_squashed, rl_module.config.action_space)

        return action

    return policy


def unpack_config(env):
    """Create a wrapped version of an env class that unpacks env_config from Ray into kwargs.

    Necessary when setting

    .. code-block:: python

        config.environment(
            env=unpack_config(SatelliteTasking),
            env_config=env_args
        )

    which generates environments that look like

    .. code-block:: python

        SatelliteTasking(**env_args)

    since RLlib expects the environment to take a dictionary called ``kwargs`` instead
    of the actual arguments.

    """

    class UnpackedEnv(env):
        def __init__(self, env_config):
            super().__init__(**env_config)

    UnpackedEnv.__name__ = f"{env.__name__}_Unpacked"

    return UnpackedEnv


def _satellite_tasking_env_creator(env_config):
    if "episode_data_callback" in env_config:
        episode_data_callback = env_config.pop("episode_data_callback")
    else:
        episode_data_callback = None
    if "satellite_data_callback" in env_config:
        satellite_data_callback = env_config.pop("satellite_data_callback")
    else:
        satellite_data_callback = None

    return EpisodeDataWrapper(
        SatelliteTasking(**env_config),
        episode_data_callback=episode_data_callback,
        satellite_data_callback=satellite_data_callback,
    )


register_env("SatelliteTasking-RLlib", _satellite_tasking_env_creator)


def _constellation_tasking_env_creator(env_config):
    if "episode_data_callback" in env_config:
        episode_data_callback = env_config.pop("episode_data_callback")
    else:
        episode_data_callback = None
    if "satellite_data_callback" in env_config:
        satellite_data_callback = env_config.pop("satellite_data_callback")
    else:
        satellite_data_callback = None

    return ParallelPettingZooEnv(
        EpisodeDataParallelWrapper(
            ConstellationTasking(**env_config),
            episode_data_callback=episode_data_callback,
            satellite_data_callback=satellite_data_callback,
        )
    )


register_env("ConstellationTasking-RLlib", _constellation_tasking_env_creator)


__doc_title__ = "RLlib Utilities"
__all__ = ["unpack_config", "load_torch_mlp_policy"]



import re
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from Basilisk.utilities.orbitalMotion import ClassicElements

from bsk_rl.utils import orbital


class TestRandomOrbit:
    @pytest.mark.repeat(10)
    def test_random_orbit(self):
        oe = orbital.random_orbit(i=None, Omega=None, omega=None, f=None)
        assert -np.pi <= oe.i <= np.pi
        assert 0 <= oe.Omega <= 2 * np.pi
        assert 0 <= oe.omega <= 2 * np.pi
        assert 0 <= oe.f <= 2 * np.pi

    @pytest.mark.repeat(10)
    def test_random_circular_orbit(self):
        oe = orbital.random_circular_orbit(i=None, Omega=None, f=None)
        assert -np.pi <= oe.i <= np.pi
        assert 0 <= oe.Omega <= 2 * np.pi
        assert 0 <= oe.f <= 2 * np.pi

    def test_repeatable(self):
        np.random.seed(0)
        oe1 = orbital.random_orbit()
        np.random.seed(0)
        oe2 = orbital.random_orbit()
        assert oe1.f == oe2.f

    def test_units(self):
        oe = orbital.random_orbit(i=90.0, a=1500, e=0.1, Omega=90.0, omega=90.0, f=90.0)
        assert oe.a == 1500000
        assert oe.e == 0.1
        assert np.pi / 2 == oe.i == oe.Omega == oe.omega == oe.f


def classic_elements(a=0, e=0, i=0, Omega=0, omega=0, f=0):
    oe = ClassicElements()
    oe.a = a
    oe.e = e
    oe.i = i
    oe.Omega = Omega
    oe.omega = omega
    oe.f = f
    return oe


class TestWalkerDeltaArgs:
    @patch(
        "bsk_rl.utils.orbital.walker_delta",
        MagicMock(return_value=[classic_elements(), classic_elements()]),
    )
    def test_randomize_ta_and_lan(self):
        sats = [MagicMock(), MagicMock()]
        walker_delta_arg_setup = orbital.walker_delta_args(
            n_planes=1, randomize_true_anomaly=True, randomize_lan=True
        )
        sat_arg_map = walker_delta_arg_setup(sats)
        assert sat_arg_map[sats[0]]["oe"].f == sat_arg_map[sats[1]]["oe"].f
        assert sat_arg_map[sats[0]]["oe"].f != 0
        assert sat_arg_map[sats[0]]["oe"].Omega == sat_arg_map[sats[1]]["oe"].Omega
        assert sat_arg_map[sats[0]]["oe"].Omega != 0

    @patch(
        "bsk_rl.utils.orbital.walker_delta",
        MagicMock(return_value=[classic_elements(), classic_elements()]),
    )
    def test_dont_randomize_ta_and_lan(self):
        sats = [MagicMock(), MagicMock()]
        walker_delta_arg_setup = orbital.walker_delta_args(
            n_planes=1, randomize_true_anomaly=False, randomize_lan=False
        )
        sat_arg_map = walker_delta_arg_setup(sats)
        assert sat_arg_map[sats[0]]["oe"].f == sat_arg_map[sats[1]]["oe"].f
        assert sat_arg_map[sats[0]]["oe"].f == 0
        assert sat_arg_map[sats[0]]["oe"].Omega == sat_arg_map[sats[1]]["oe"].Omega
        assert sat_arg_map[sats[0]]["oe"].Omega == 0


class TestRelativeToChief:
    def test_relative_to_chief(self):
        chief_name = "ChiefSat"
        chief_orbit = classic_elements(a=1)
        deputy_relative_state = {
            "DeputySat1": np.zeros(6),
            "DeputySat2": np.array([1, 0, 0, 0, 0, 0]),
        }

        relative_orbit_arg_setup = orbital.relative_to_chief(
            chief_name, chief_orbit, deputy_relative_state
        )

        satellites = [
            MagicMock(sat_args_generator=dict(mu=1)),
            MagicMock(),
            MagicMock(),
        ]
        satellites[0].name = "ChiefSat"
        satellites[1].name = "DeputySat1"
        satellites[2].name = "DeputySat2"
        args = relative_orbit_arg_setup(satellites)
        assert chief_orbit.a == args[satellites[0]]["oe"].a
        assert chief_orbit.a == args[satellites[1]]["oe"].a
        assert chief_orbit.e == args[satellites[0]]["oe"].e
        assert chief_orbit.e == args[satellites[1]]["oe"].e
        assert chief_orbit.i == args[satellites[0]]["oe"].i
        assert chief_orbit.i == args[satellites[1]]["oe"].i

        assert np.isclose(args[satellites[2]]["oe"].a, -1 / 3)
        assert np.isclose(args[satellites[2]]["oe"].e, 7.0)
        assert np.isclose(args[satellites[2]]["oe"].i, 0.0)


class TestRandomEpoch:
    @pytest.mark.repeat(10)
    def test_random_epoch(self):
        assert (
            re.match(
                r"\d{4} [A-Z]{3} \d{2} \d{2}:\d{2}:\d{2}\.\d{3} \(UTC\)",
                orbital.random_epoch(),
            )
            is not None
        )

    def test_repeatable(self):
        np.random.seed(0)
        e1 = orbital.random_epoch()
        np.random.seed(0)
        e2 = orbital.random_epoch()
        assert e1 == e2


@pytest.mark.parametrize(
    "r_sat,r_target,expected",
    [
        (np.array([2, 0, 0]), np.array([1, 0, 0]), np.pi / 2),
        (
            np.array([[1, 1, 0], [2, 0, 0]]),
            np.array([1, 0, 0]),
            np.array([0, np.pi / 2]),
        ),
    ],
)
def test_elevation(r_sat, r_target, expected):
    assert np.all(orbital.elevation(r_sat, r_target) == expected)


class TestTrajectorySimulator:
    epoch = "2005 JUL 24 20:50:33.771 (UTC)"
    oe = orbital.random_orbit(i=0.0, omega=0, Omega=0, f=45.0)
    mu = 0.3986004415e15

    @pytest.mark.parametrize(
        "kwargs,error",
        [
            (dict(rN=[1, 0, 0], vN=[0, 1, 0]), False),
            (dict(oe=oe, mu=mu), False),
            (dict(rN=[1, 0, 0], oe=oe), True),
            (dict(rN=[1, 0, 0], oe=oe, mu=mu), True),
            (dict(rN=[1, 0, 0], vN=[0, 1, 0], oe=oe), True),
            (dict(), True),
        ],
    )
    def test_init(self, kwargs, error):
        if error:
            with pytest.raises(ValueError):
                orbital.TrajectorySimulator(self.epoch, **kwargs)
        else:
            orbital.TrajectorySimulator(self.epoch, **kwargs)

    @pytest.mark.parametrize(
        "dt,extend1,extend2",
        [(30.0, 100.0, 200.0), (40.0, 100.0, 50.0)],
    )
    def test_extend_to_and_time_properties(self, dt, extend1, extend2):
        ts = orbital.TrajectorySimulator(self.epoch, oe=self.oe, mu=self.mu, dt=dt)
        assert ts.sim_time == 0
        ts.extend_to(extend1)
        assert ts.sim_time == dt * np.floor(extend1 / dt)
        np.testing.assert_allclose(ts.times, np.arange(0, extend1, dt), atol=1e-9)
        ts.extend_to(extend2)
        assert ts.sim_time == dt * np.floor(max(extend1, extend2) / dt)
        np.testing.assert_allclose(
            ts.times, np.arange(0, max(extend1, extend2), dt), atol=1e-9
        )

    def test_eclipse(self):
        ts = orbital.TrajectorySimulator(self.epoch, oe=self.oe, mu=self.mu, dt=500.0)
        start_1, end_1 = ts.next_eclipse(0.0)
        # Verify eclipse start and end are correct (duration is less than 1/2 orbit)
        assert (end_1 - start_1 % 5700) < 5700 / 2
        # With chosen params, start out of eclipse
        assert end_1 > start_1
        # Go into eclipse
        start_2, end_2 = ts.next_eclipse(start_1 + 1.0)
        assert end_2 == end_1  # Soonest close is the same
        assert abs(start_2 - start_1 - 5700) < 500.0  # Starts are about an orbit apart

    def test_no_eclipse(self):
        ts = orbital.TrajectorySimulator(
            self.epoch,
            oe=orbital.random_orbit(a=50000, i=90.0, omega=0, Omega=0, f=45.0),
            mu=self.mu,
        )
        assert ts.next_eclipse(0, max_tries=3) == (1.0, 1.0)

    def test_interpolators(self):
        # Weak tests, could be better
        ts = orbital.TrajectorySimulator(self.epoch, oe=self.oe, mu=self.mu)
        assert (ts.r_BN_N(0) == ts.rN_init).all()
        assert (ts.r_BN_N(0) != ts.r_BN_N(1)).all()
        ts = orbital.TrajectorySimulator(self.epoch, oe=self.oe, mu=self.mu)
        assert (ts.r_BP_P(0) != ts.r_BP_P(1)).all()


class TestFunctions:
    def test_rv2HN(self):
        r = np.array([1, 0, 0])
        v = np.array([0, 1, 0])
        HN = orbital.rv2HN(r, v)
        assert np.allclose(HN, np.eye(3))

    def test_rv2HN_nonunit(self):
        r = np.array([1, 0, 0]) * 12345
        v = np.array([0, 1, 0]) * 678.9
        HN = orbital.rv2HN(r, v)
        assert np.allclose(HN, np.eye(3))

    def test_rv2omega(self):
        r = np.array([1, 0, 0])
        v = np.array([0, 1, 0])
        omega = orbital.rv2omega(r, v)
        assert np.allclose(omega, np.array([0, 0, 1]))

    @pytest.mark.repeat(10)
    def test_cd_hill(self):
        rc_N = np.random.rand(3)
        vc_N = np.random.rand(3)
        rd_N = np.random.rand(3)
        vd_N = np.random.rand(3)

        rho_H, rho_deriv_H = orbital.cd2hill(rc_N, vc_N, rd_N, vd_N)
        rd_N_new, vd_N_new = orbital.hill2cd(rc_N, vc_N, rho_H, rho_deriv_H)
        assert np.allclose(rd_N, rd_N_new)
        assert np.allclose(vd_N, vd_N_new)
